/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   xecline.c
 *    freem interpreter proper
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdlib.h>

#include "mpsdef.h"
#include "namespace.h"
#include "transact.h"
#include "merge.h"
#include "sighnd.h"
#include "init.h"
#include "events.h"
#include "mdebug.h"
#include "freem.h"
#include "mref.h"
#include "log.h"
#include "io_socket.h"
#include "merr.h"

#ifdef __EMSCRIPTEN__
#include <emscripten.h>
#include <emscripten/html5.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <stdio.h>
#include <ctype.h>

#include <time.h>
#include <sys/types.h>
#if !defined(__OpenBSD__) && !defined(__FreeBSD__)
# include <sys/timeb.h>
#endif
#include <sys/wait.h>
#include <sys/time.h>

#if defined(HAVE_READLINE_READLINE_H)
# include <readline/readline.h>
#elif defined(HAVE_READLINE_H)
# include <readline.h>
#endif

#if defined(HAVE_READLINE_HISTORY_H)
# include <readline/history.h>
#elif defined(HAVE_HISTORY_H)
# include <history.h>
#endif

/* mumps commands */
#define ABLOCK       0
#define ASSIGN      'a'
#define ASTART       1
#define ASTOP        2
#define AUNBLOCK     3
#define BREAK       'b'
#define CLOSE       'c'
#define DO          'd'
#define DO_BLOCK     4
#define ELSE        'e'
#define ESTART       5
#define ESTOP        6
#define ETRIGGER     7
#define FOR         'f'     /* ALSO in symtab.c */
#define GOTO        'g'
#define HA          'h'
#define HALT         8
#define HANG         9
#define IF          'i'
#define JOB         'j'
#define KILL        'k'
#define KVALUE      10
#define KSUBSC      11
#define LOCK        'l'
#define MERGE       'm'
#define NEWCMD      'n'
#define OPEN        'o'
#define QUIT        'q'
#define READ        'r'
#define RLOAD       12
#define RSAVE       13
#define SET         's'
#define TCOMMIT     14
#define THEN        15
#define TRESTART    16
#define TROLLBACK   17
#define TSTART      18
#define USE         'u'
#define VIEW        'v'
#define WRITE       'w'
#define XECUTE      'x'

#define ZALLOCATE   'A'
#define ZBREAK      'B'
#define ZDEALLOCATE 'D'
#define ZGO         'G'
#define ZHALT       'H'
#define ZINSERT     'I'
#define ZJOB        'J'
#define ZLOAD       'L'
#define ZNAMESPACE  'N' // formerly ZNEW
#define ZPRINT      'P'
#define ZQUIT       'Q'
#define ZREMOVE     'R'
#define ZSAVE       'S'
#define ZTRAP       'T'
#define ZWRITE      'W'
#define ZWATCH      20
#define ZASSERT     21
#define PRIVATE     SP

void on_frame_entry(void);

/*
 * xecline():
 *   typ (where to go on function entry):   1 = restart
 *                                          2 = error
 *
 */
int xecline(int typ)
{
    short libcall = FALSE;
    char *namold;
    long rouoldc;
    unsigned long jobtime;
    char label[256], routine[256];
    char vn[65535], tmp[256], tmp2[256], tmp3[256];

    char *deferrable_codptr;
    char deferrable_code[512];
    
    register int i;
    register int j;
    register int ch;
    
#   ifdef DEBUG_NEWSTACK
    int loop;
#   endif

    deferrable_codptr = deferrable_code;

    if (typ == 1) goto restart;
    if (typ == 2) goto err;
    if (typ == 3) {
        libcall = TRUE;
        goto restart;
    }

    /*      if (typ==0) goto next_line; */

next_line:          /* entry point for next command line */

    while ((roucur < rouend) && (ch = (*roucur++)) != TAB && ch != SP); /* skip label */

    if (roucur >= rouend) goto quit0;         /* end of routine implies QUIT */

    while ((ch = *roucur) == TAB || ch == SP) roucur++;

    i = 0;
    if (ch == '.') {               /* get level count */
        
        do {
            i++;
            while ((ch = (*++roucur)) == SP || ch == TAB);
        }

        while (ch == '.');
    }

    if (i != level) {

        if (mcmnd == GOTO) {
            ierr = M45;
            goto err;
        }

        if (i < level) {
            goto quit0;
        }
        else {
            roucur += stlen (roucur) + 2;
            goto next_line;
        }
    }

    i = stcpy(code, roucur) + 1;    
    code[i] = EOL;
    roucur += i + 1;
    codptr = code;

next_cmnd:          /* continue line entry point */


    if (evt_async_enabled == TRUE) {

        /* process async events */
        char *ev_handlers = (char *) malloc (STRLEN * sizeof (char));


        /* get a comma-delimited list of applicable handlers (e.g. ^HNDL1,^HNDL2,^HNDL3) */
        ev_handlers[0] = NUL;
        evt_depth = evt_get_handlers (ev_handlers);

        stcnv_c2m (ev_handlers);
        stcpy (tmp3, ev_handlers);
        free (ev_handlers);

        /* only execute event handlers if we have at least one such handler registered in ^$JOB($JOB,"EVENTS") */
        if (evt_depth) {
    
            /* per the event extension, async events are to be disabled during the execution of event handlers */

            /* TODO: this should be done by incrementing the event block counter
               for all event types, or whatever the event extension says to do.

               In any event (rimshot here for the obvious pun), turning off all
               event handlers this way is decidedly non-standard. Or non-what-might-
               become the standard. Whatever. */
            
            evt_async_enabled = FALSE;
            evt_async_initial = TRUE;
            evt_async_restore = TRUE;

            goto evthandler;
    
        }

    }


    if (ierr > OK) goto err;

next0:

        do {
            if ((ch = *codptr) == EOL) {
            if (forsw) goto for_end;

            goto next_line;
        }

        codptr++;
    }

    while (ch == SP);

    /* decode command word */

    if (ch < 'A') {  /* Handle non-alpha first chars */
        
        if (ch == ';') {           /* COMMENT */

            ch = *(codptr++);

            if(ch == '%') {         /* DIRECTIVE */

                int dir_pos = 0;
                int dir_wc = 0;
                char dir_words[20][255];

                while((ch = *(codptr++)) != EOL) {

                    switch (ch) {


                        case SP:
                            
                            dir_words[dir_wc][dir_pos] = NUL;

                            dir_wc++;
                            dir_pos = 0;

                            break;


                        default:

                            dir_words[dir_wc][dir_pos++] = ch;

                    }

                }

                dir_words[dir_wc][dir_pos] = NUL;

                if(strcmp(dir_words[0], "MODE") == 0) {
                    if(strcmp(dir_words[1], "STRICT") == 0) {
                        strict_mode = 1;
                    }
                    else if(strcmp(dir_words[1], "LEGACY") == 0) {
                        strict_mode = 0;
                    }
                    else {
                        ierr = CMMND;
                        goto err;
                    }
                    goto skip_line;
                }
                else {
                    goto skip_line;
                }

            }

            goto skip_line;
        }

        if ((!standard) && (ch == '#')) {
            goto skip_line;
        }

        if ((standard) && (ch == '#')) {
            ierr = NOSTAND;
            goto err;
        }

        if (ch == '!') {           /* UNIXCALL */
            
            if (standard) {
                ierr = NOSTAND;
                goto err;
            }

            close_all_globals ();   /* close all globals */
            set_io (UNIX);

            /* don't catch child dies signal */
            sig_attach (SIGUSR1, SIG_IGN);

            tmp2[stcpy (tmp2, codptr)] = NUL;

            if (demomode) fputc (d1char, stdout);

            if (tmp2[0] == '!') {           /* call does not return!! *//* close all files, except for stdin etc. */
                
                /* that's not done automatically          */
                j = 1;
                while (j <= MAXDEV) {
                    
                    if (jour_flag && (j == 2)) {
                        j++;
                        continue;
                    }

                    if (devopen[j]) fclose (opnfile[j]);

                    devopen[j++] = 0;
                }

                io = HOME;
                lock (" \201", -1, LOCK);   /* un-LOCK */
                lock (" \201", -1, ZDEALLOCATE);    /* ZDEALLOCATE ALL */

#if !defined(__AMIGA)                
                zsystem = execl ("/bin/sh", "sh", "-c", &tmp2[1], 0, NULL);
#endif            
            }

#if !defined(EMSCRIPTEN)

            else if (tmp2[0] == '<') {           /* call write output to %-array */
                
                FILE *pipdes;
                char key[STRLEN + 1 /*was 256 */ ];
                char data[STRLEN + 1 /*was 256 */ ];

                vn[0] = '%';
                vn[1] = EOL;
                
                symtab (kill_sym, vn, data);
                
                vn[0] = '%';
                vn[1] = EOL;
                data[0] = '0';
                data[1] = EOL;
                
                symtab (set_sym, vn, data);

                if ((pipdes = popen (&tmp2[1], "r")) == NULL) {
                    zsystem = 1;
                }
                else {
                    
                    while (fgets (data, STRLEN, pipdes)) {
                        vn[0] = '%';
                        vn[1] = EOL;
                        key[0] = '%';
                        key[1] = DELIM;
                        
                        symtab (getinc, vn, &key[2]);
                        
                        i = strlen (data);
                        
                        data[i] = EOL;

                        if (i > 1 && data[i - 1] == LF) data[i - 1] = EOL;

                        symtab (set_sym, key, data);
                        
                        if (ierr == STORE) break;
                    }

                    pclose (pipdes);
                    
                    zsystem = 0;
                }
            }      
            else if (tmp2[0] == '>') {           /* call read input from %-array */
                FILE *pipdes;
                char key[STRLEN + 1 /*was 256 */ ];
                char data[STRLEN + 1 /*was 256 */ ];
                int i, k, l;

                vn[0] = '%';
                vn[1] = EOL;
                
                symtab (get_sym, vn, data);

                ierr = OK;
                k = intexpr (data);

                if (k < 1 || (pipdes = popen (&tmp2[1], "w")) == NULL) {
                    zsystem = 1;
                }
                else {
                    
                    for (i = 1; i <= k; i++) {
                        key[0] = '%';
                        key[1] = DELIM;
                        
                        intstr (&key[2], i);                        
                        symtab (get_sym, key, data);

                        l = stlen (data);
                        
                        data[l++] = LF;
                        data[l] = NUL;

                        fputs (data, pipdes);
                    }
                    
                    pclose (pipdes);

                    zsystem = 0;
                    ierr = OK;
                }
            }
#endif            
            else {
                zsystem = system (tmp2);
            }

            if (demomode) fputc (d1char, stdout);

            sig_attach (SIGUSR1, &oncld);   /* restore handler */

            set_io (MUMPS);

            if (ierr == STORE) {
                zsystem = 1;
                goto err;
            }

            goto skip_line;
        }

        ierr = CMMND;
        goto err;
    } /* END handling of non-alpha first chars */


    mcmnd = ch | 0140;      /* uppercase to lower case */

    i = 1;

    while ((ch = (*codptr)) != SP && ch != ':' && ch != EOL) {
        tmp3[++i] = ch | 0140;
        codptr++;
    }

    j = i;

    if (j > 1) {

        tmp3[0] = SP;
        tmp3[1] = mcmnd;
        tmp3[++j] = SP;
        tmp3[++j] = EOL;
        
        if (mcmnd != 'z') {
            
            if (find (" ab ablock assign asta astart asto astop aunb aunblock \
            break close do else esta estart esto estop etr etrigger for goto hang halt \
            if job kill ks ksubscripts kv kvalue lock merge new open quit read rl rload\
            rs rsave set tc tcommit th then tre trestart tro trollback ts tstart use\
            view write xecute ", tmp3) == FALSE) {
                ierr = CMMND;
                goto err;
            }

            switch (mcmnd)
            {

            case 'a':
                if (tmp3[2] == 'b') mcmnd = ABLOCK;
                else if (tmp3[2] == 'u') mcmnd = AUNBLOCK;
                else if (tmp3[4] == 'a') mcmnd = ASTART;
                else if (tmp3[4] == 'o') mcmnd = ASTOP;
                break;

            case 'e':
                if (tmp3[2] == 't') mcmnd = ETRIGGER;
                else if (tmp3[4] == 'a') mcmnd = ESTART;
                else if (tmp3[4] == 'o') mcmnd = ESTOP;
                break;

            case 'h':
                mcmnd = tmp3[4] == 't' ? HALT : HANG;
                break;

            case 'k':
                if (tmp3[2] == 'v') mcmnd = KVALUE;
                else if (tmp3[2] == 's') mcmnd = KSUBSC;
                break;

            case 'm':
                /* save off pre-parse codptr and code for commands implemented in M routines */
                deferrable_codptr = codptr;
                stcpy (deferrable_code, "MERGE\201");
                stcat (deferrable_code, deferrable_codptr);

                break;

            case 'r':
                if (tmp3[2] == 'l') mcmnd = RLOAD;
                else if (tmp3[2] == 's') mcmnd = RSAVE;
                break;

            case 't':
                if (tmp3[2] == 'c') mcmnd = TCOMMIT;
                else if (tmp3[2] == 'h') mcmnd = THEN;
                else if (tmp3[3] == 'e') mcmnd = TRESTART;
                else if (tmp3[3] == 'o') mcmnd = TROLLBACK;
                else if (tmp3[2] == 's') mcmnd = TSTART;                
            }           /* end of switch(mcmnd) */
        
        }
        else {

            if (standard) {
                ierr = NOSTAND;
                goto err;
            }

            mcmnd = tmp3[2] - 32;   /* z_command select */


            if (mcmnd == 'A' && (tmp3[3] == 's' || tmp3[3] == 'S')) mcmnd = ZASSERT;
            if (mcmnd == 'W' && (tmp3[3] == 'a' || tmp3[3] == 'A')) mcmnd = ZWATCH;
            

            if (find (zcommds, tmp3) == FALSE) mcmnd = PRIVATE;
        }
    }


    if (*codptr == ':') {               

        /* handle postconditional */
        /* postcond after FOR,IF,ELSE not allowed */ 
        
        if (mcmnd == FOR || mcmnd == IF || mcmnd == ELSE) {
            ierr = SPACER;
            goto err;
        }

        codptr++;

        expr (STRING);

        if (ierr > OK) goto err;

        ch = *codptr;

        if (ch != SP && ch != EOL) {
            ierr = SPACER;
            goto err;
        }

        if (tvexpr (argptr) == FALSE) {           /* skip arguments */
            mcmnd = 0;      /* avoid false LEVEL error */
            
            for (;;) {
                if (ch == EOL) goto skip_line;
                if ((ch = *++codptr) == SP) goto next_cmnd;
                if (ch != '"') continue;

                while (*codptr++ != EOL) {
                    if (*codptr != ch) continue;
                    if (*++codptr != ch) break;
                }

                if (--codptr == code) goto err;
            }
        }

    }

    if (*codptr != EOL) {               /* beware argumentless cmnds at end of line */
        codptr++;           /* entry for next argument in list */

again:
        while (*codptr == '@') {           /* handle indirection */
            
            stcpy (tmp, codptr++);  /* save code to restore on nameind */
            expr (ARGIND);

            if (ierr > OK) goto err;

            if (((ch = (*codptr)) != SP && ch != EOL && ch != ',' && ch != ':' && ch != '=') || (ch == '@' && *(codptr + 1) == '(')) {
                stcpy (code, tmp);  /* restore code on nameind */
                codptr = code;

                break;
            }
            else {
                stcpy (argptr + stlen (argptr), codptr);
                stcpy (code, argptr);
                
                codptr = code;
            }
        }
    }

    switch (mcmnd) {

        case KVALUE:
            
            /* argumentless: KVALUE nukleurrr winturr */
            if (((ch = *codptr) == SP) || ch == EOL) {
                write_m ("Argumentless KVALUE not yet implemented.\201");
                goto next_cmnd;                
            }


            if (ch != '(') { /* inclusive KVALUE */

                for (;;) {
                    expr (NAME);                            /* try to interpret an mname */
                    
                    if (ierr > OK) goto err;

                    stcpy (vn, varnam);

                    if (vn[0] != '^') {
                        symtab (killone, vn, NULL);
                    }
                    else {
                        if (vn[1] == '$') {
                            ssvn (killone, vn, NULL);
                        }
                        else {
                            global (killone, vn, NULL);
                        }
                    }

                    if (ierr > OK) goto err;

                    if ((ch = *(codptr + 1)) == EOL) {
                        codptr++;
                        break;
                    }
                    else if (ch == ',') {
                        codptr += 2;
                    }
                    else {
                        ierr = CMMND;
                        goto err;
                    }
                } 

            }
            else { /* exclusive KVALUE */
                
                write_m ("Exclusive KVALUE not yet implemented.\201");

                while ((ch = *(codptr++)) != SP && ch != EOL) ; /* skip to the end of the command */

            }

            goto next_cmnd;

        case KSUBSC:
            {
                char *old_value;

                if ((old_value = (char *) malloc (STRLEN * sizeof (char))) == NULL) {
                    ierr = MEMOV;
                    goto err;
                }

                /* argumentless: KSUBSCRIPTS nukleurrr winturr */
                if (((ch = *codptr) == SP) || ch == EOL) {
                    write_m ("Argumentless KSUBSCRIPTS not yet implemented.\201");

                    free (old_value);

                    goto next_cmnd;                
                }


                if (ch != '(') { /* inclusive KSUBSCRIPTS */

                    for (;;) {
                        expr (NAME);                            /* try to interpret an mname */
                        
                        if (ierr > OK) goto err;

                        stcpy (vn, varnam);

                        if (vn[0] != '^') {
                            symtab (get_sym, vn, old_value);
                            symtab (kill_sym, vn, NULL);
                            symtab (set_sym, vn, old_value);
                        }
                        else {
                            if (vn[1] == '$') {
                                ssvn (get_sym, vn, old_value);
                                ssvn (kill_sym, vn, NULL);
                                ssvn (set_sym, vn, old_value);
                            }
                            else {
                                global (get_sym, vn, old_value);
                                global (kill_sym, vn, NULL);
                                global (set_sym, vn, old_value);
                            }
                        }

                        free (old_value);

                        if (ierr > OK) goto err;

                        if ((ch = *(codptr + 1)) == EOL) {
                            codptr++;
                            break;
                        }
                        else if (ch == ',') {
                            codptr += 2;
                        }
                        else {
                            ierr = CMMND;
                            goto err;
                        }
                    } 

                }
                else { /* exclusive KSUBSCRIPTS */
                    
                    write_m ("Exclusive KSUBSCRIPTS not yet implemented.\201");

                    while ((ch = *(codptr++)) != SP && ch != EOL) ; /* skip to the end of the command */

                }

                goto next_cmnd;
            }


        case TSTART:
            tp_tstart ("foo", FALSE, FALSE, NULL);

            goto next_cmnd;
            
            break;
        case TCOMMIT:
            tp_tcommit ();

            goto next_cmnd;

            break;
        case TRESTART:
            tp_trestart ();

            goto next_cmnd;

            break;
        case TROLLBACK:
            tp_trollback (tp_level);

            break;
        case SET:

set0:
            if ((ch = (*codptr)) >= 'A') {           /* no set$piece nor multiset */

                if (*(codptr + 1) == '=' && *(codptr + 2) != '=') { /* single char local variable */
                    
                    if ((ch > 'Z' && ch < 'a') || ch > 'z') {
                        ierr = INVREF;
                        break;
                    }

                    vn[0] = ch;
                    vn[1] = EOL;
                    codptr += 2;

                    expr (STRING);

                    if (ierr > OK) break;

                    symtab (set_sym, vn, argptr);
                    if (ierr > OK) break;

                    goto set1;

                }

                expr (NAME);

                if (ierr > OK) break;

                stcpy (vn, varnam);

                if ((*++codptr != '=') || (*(codptr + 1) == '=')) {
                    ch = *codptr;
                    
                    /* double char symbol ** (power) is encoded by ' ' */
                    if (ch == '*' && *(codptr + 1) == ch) {
                        codptr++;
                        ch = ' ';
                    }

                    /* negated boolean operator */
                    else if ((ch == '\'') && (*(codptr + 2) == '=')) ch = SETBIT (*++codptr);

                    if (*++codptr != '=') {

/* SET A++ or SET A-- equivalent to SET A+=1 SET A-=1 currently disabled */
//#ifdef NEVER
                        if ((ch == '+' || ch == '-') && ch == *codptr) {
                            codptr++;
                            setop = ch;
                            argptr[0] = '1';
                            argptr[1] = EOL;
                            
                            goto set2;
                        }
//#endif /* NEVER */

                        ierr = ASSIGNER;
                        break;
                    }

                    setop = ch;
                }

                codptr++;

                expr (STRING);

                if (ierr > OK) break;

set2:
                if (vn[0] == '^') {
                    
                    if (vn[1] == '$') {
                        ssvn (set_sym, vn, argptr);
                    }
                    else {
                        global (set_sym, vn, argptr);
                    }

                }
                else {
                    symtab (set_sym, vn, argptr);                
                }

                if (ierr > OK) {
                    stcpy (varerr, vn);
                    break;
                }
set1:
                if (*codptr != ',') break;

                if (*++codptr == '@') goto again;

                goto set0;
            }

            /****** special SET syntax: multiple SET, set$piece, special variables */
            {
                char multiset, vnset[256];  /* multiset variables */
                long arg3, arg4;    /* 3rd,4th arg in set$piece */

                if ((multiset = (ch == '('))) {
                    vnset[0] = EOL;
                    codptr++;
                }

set:
                if (*codptr == '$' && (*(codptr + 1) | 0140) == 'p') {           /* set$piece */

                    if (multiset) {
                        ierr = INVREF;
                        goto err;
                    }
        
                    setpiece = 'p';

                    while (*++codptr != '(') {
                        
                        if (*codptr == EOL) {
                            ierr = INVREF;
                            goto err;
                        }
                    
                    }
                    
                    codptr++;
                    
                    expr (NAME);

                    if (ierr > OK) goto err;

                    stcpy (vn, varnam);
                    
                    codptr++;

                    if (*codptr++ != ',') {
                        ierr = COMMAER;
                        goto err;
                    }

                    expr (STRING);

                    if (ierr > OK) goto err;

                    stcpy (tmp2, argptr);

                    if (*codptr != ')') {
                    
                        codptr++;

                        expr (STRING);

                        if (ierr > OK) goto err;

                        arg3 = intexpr (argptr);

                        if (ierr == MXNUM) {
                            arg3 = 256;
                            ierr = OK;
                        }
                    
                    }
                    else {
                        arg3 = 1;
                    }

                    if (*codptr != ')') {
                    
                        codptr++;

                        expr (STRING);

                        if (ierr > OK) goto err;

                        if (*codptr != ')') {
                            ierr = BRAER;
                            goto err;
                        }

                        arg4 = intexpr (argptr);

                        if (ierr == MXNUM) {
                            arg4 = 256;
                            ierr = OK;
                        }
                    
                    }
                    else {
                        arg4 = arg3;
                    }
                } /* set$piece */
                else if (*codptr == '$' && (*(codptr + 1) | 0140) == 'q' && (*(codptr + 2) | 0140) == 's') { /* TODO: verify this works (jpw) was (*codptr == '$q'...*/
                    /*SET $QSUBSCRIPT */

                    if (multiset) {
                        ierr = INVREF;
                        goto err;
                    }

                    setpiece = 'q';

                    while (*++codptr != '(') {
                        
                        if (*codptr == EOL) {
                            ierr = INVREF;
                            goto err;
                        }

                    }
                    
                    codptr++;

                    expr (NAME);

                    if (ierr > OK) goto err;

                    stcpy (vn, varnam);

                    if (*++codptr == ',') {
                        codptr++;

                        expr (STRING);
                        
                        if (ierr > OK) goto err;

                        stcpy (tmp2, argptr);
                    }

                    if (*codptr != ')') {
                        ierr = BRAER;
                        goto err;
                    }
                
                }
                else if (*codptr == '$' && (*(codptr + 1) | 0140) == 'e' && (*(codptr + 2) | 0140) != 't' && (*(codptr + 2) | 0140) != 'c') {

                    /* set $extract */
                    if (multiset) {
                        ierr = INVREF;
                        goto err;
                    }

                    setpiece = 'e';
                    
                    while (*++codptr != '(') {
                        
                        if (*codptr == EOL) {
                            ierr = INVREF;
                            goto err;
                        }

                    }

                    codptr++;

                    expr (NAME);

                    if (ierr > OK) goto err;

                    stcpy (vn, varnam);

                    codptr++;

                    if (*codptr != ')') {
                        codptr++;

                        expr (STRING);
                        
                        if (ierr > OK) goto err;

                        arg3 = intexpr (argptr);
                        
                        if (ierr == MXNUM) {
                            arg3 = 256;
                            ierr = OK;
                        }
                    }
                    else {
                        arg3 = 1;
                    }

                    if (*codptr != ')') {
                        codptr++;

                        expr (STRING);
                        
                        if (ierr > OK) goto err;

                        if (*codptr != ')') {
                            ierr = BRAER;
                            goto err;
                        }

                        arg4 = intexpr (argptr);

                        if (ierr == MXNUM) {
                            arg4 = 256;
                            ierr = OK;
                        }

                    }
                    else {
                        arg4 = arg3;
                    }

                }
                else if (*codptr == '$' && (*(codptr + 1) | 0140) == 'e' && (*(codptr + 2) | 0140) == 'c') {
                    /* set $ecode */

                    if (multiset) {
                        ierr = INVREF;
                        goto err;
                    }

                    while ((*(++codptr)) != '=');

                    codptr++;

                    expr (STRING);

                    if (ierr > OK) goto err;

                    ierr = merr_set_ecode (argptr);

#if 0
                    set_io (UNIX);
                    stcnv_m2c (ecode);
                    stcnv_m2c (etrap);
                    printf ("\n\n*** IN SET $ECODE: ecode = '%s' etrap = '%s'\n", ecode, etrap);
                    stcnv_c2m (etrap);
                    stcnv_c2m (ecode);
                    set_io (MUMPS);
#endif

                    if (ierr > OK) goto err;

                    goto s_end;

                }
                else if (*codptr == '$' && (*(codptr + 1) | 0140) == 'e' && (*(codptr + 2) | 0140) == 't') {
                    /* set $etrap */

                    if (multiset) {
                        ierr = INVREF;
                        goto err;
                    }

                    while ((*(++codptr)) != '=');

                    codptr++;

                    expr (STRING);

                    if (ierr > OK) goto err;

                    stcpy (etrap, argptr);

#if 0
                    set_io (UNIX);
                    stcnv_m2c (ecode);
                    stcnv_m2c (etrap);
                    printf ("\n\n***IN SET $ETRAP: ecode = '%s' etrap = '%s'\n", ecode, etrap);
                    stcnv_c2m (etrap);
                    stcnv_c2m (ecode);
                    set_io (MUMPS);
#endif

                    goto s_end;

                }
                else if (*codptr == '$' && (*(codptr + 1) | 0140) == 'g' && cset) {           /* set$get */
                    
                    if (multiset) {
                        ierr = INVREF;
                        goto err;
                    }

                    setpiece = 'g';

                    while (*++codptr != '(') {
                        
                        if (*codptr == EOL) {
                            ierr = INVREF;
                            goto err;
                        }
                    }

                    codptr++;

                    expr (NAME);

                    if (ierr > OK) goto err;

                    stcpy (vn, varnam);

                    if (*++codptr == ',') {
                        codptr++;

                        expr (STRING);

                        if (ierr > OK) goto err;

                        stcpy (tmp2, argptr);
                    }
                    else {
                        tmp2[0] = EOL;
                    }

                    if (*codptr != ')') {
                        ierr = BRAER;
                        goto err;
                    }
                }
                else {
                    if (*codptr == '$') {
                        codptr++;

                        expr (NAME);

                        if (ierr > OK) goto err;

                        stcpy (tmp, varnam);

                        varnam[0] = '$';

                        stcpy (&varnam[1], tmp);
                        
                        i = 0;
                        while ((ch = varnam[++i]) != EOL) { 
                            if (ch >= 'A' && ch <= 'Z') {
                                varnam[i] |= 0140;  /*to lowercase */
                            }
                        }
                    }
                    else {
                        expr (NAME);

                        if (ierr > OK) goto err;
                    }

                    stcpy (vn, varnam);
                }

                if (multiset) {
                    vnset[i = stlen (vnset)] = SOH;
                    
                    stcpy (&vnset[++i], vn);

                    if (*++codptr == ',') {
                        codptr++;
                        goto set;
                    }

                    if (*codptr != ')') {
                        ierr = COMMAER;
                        goto err;
                    }
                }

                if (*++codptr != '=') {
                    ch = *codptr;

                    if (!cset || *++codptr != '=' || multiset || setpiece || varnam[0] == '$') {
                        ierr = ASSIGNER;
                        break;
                    }

                    setop = ch;
                }

                codptr++;

                expr (STRING);

                if (ierr > OK) goto err;

                if (multiset)
multi:
                {
                    i = 0;
                    while (vnset[i] == SOH) i++;

                    j = 0;
                    while ((vn[j] = vnset[i]) != SOH && vnset[i] != EOL) {
                        vnset[i++] = SOH;
                        j++;
                    }

                    vn[j] = EOL;

                    if (j == 0) goto s_end;
                }

                if (setpiece == 'p') {
                    short m, n;

                    if (arg4 < arg3 || arg4 < 1) {
                        setpiece = FALSE;
                        break;
                    }

                    if (arg3 <= 0) arg3 = 1;
                    
                    if (vn[0] == '^') {
                        if (vn[1] == '$') {
                            ssvn (get_sym, vn, tmp3);
                        }
                        else {
                            global (get_sym, vn, tmp3);
                        }
                    }
                    else {
                        symtab (get_sym, vn, tmp3);
                    }


                    if (ierr == UNDEF || ierr == (UNDEF - CTRLB) || ierr == M6 || ierr == (M6 - CTRLB) || ierr == M7 || ierr == (M7 - CTRLB)) {
                        tmp3[0] = EOL;
                        ierr = OK;
                    }
                    else if (ierr != OK) {
                        stcpy (varerr, vn);
                    }

                    ch = 0;
                    m = 0;
                    n = 0;

                    j = stlen (tmp2);

                    while (n < arg3 - 1) {
                        
                        if ((ch = find (&tmp3[m], tmp2)) <= 0) {

                            while (++n < arg3) {
                                if (stcat (tmp3, tmp2) == 0) {
                                    ierr = MXSTR;
                                    goto err;
                                }
                            }

                            arg3 = arg4 = stlen (tmp3);

                            goto set10;
                        }

                        n++;
                        m += j + ch - 1;
                    }

                    if (arg3 > 1) {
                        arg3 = m;
                    }
                    else {
                        arg3 = 0;
                    }

                    while (n++ < arg4) {
                        
                        if ((ch = find (&tmp3[m], tmp2)) <= 0) {
                            arg4 = stlen (tmp3);

                            goto set10;
                        }

                        m += j + ch - 1;
                    }

                    arg4 = m - j;

set10:

                    stcpy0 (tmp2, tmp3, (long) arg3);
                    
                    tmp2[arg3] = EOL;

                    if (stcat (tmp2, argptr) == 0) {
                        ierr = MXSTR;
                        goto err;
                    }

                    if (stcat (tmp2, &tmp3[arg4]) == 0) {
                        ierr = M56; //JPW
                        goto err;
                    }

                    stcpy (argptr, tmp2);
                    
                    setpiece = FALSE;
                }
                else if (setpiece == 'q') {           /* SET$QSUBSCRIPT */
                    
                    setpiece = FALSE;
                    
                    if (vn[0] == '^') {
                        if (vn[1] == '$') {
                            ssvn (get_sym, vn, tmp3);
                        }
                        else {
                            global (get_sym, vn, tmp3);
                        }                        
                    }
                    else {
                        symtab (get_sym, vn, tmp3);
                    }

                    if (ierr == UNDEF || ierr == (UNDEF - CTRLB) || ierr == M6 || ierr == (M6 - CTRLB) || ierr == M7 || ierr == (M7 - CTRLB)) {
                        tmp3[0] = EOL;
                        ierr = OK;
                    }
                    else if (ierr != OK) {
                        stcpy (varerr, vn);
                    }

                    if (ierr == OK) {
                        /* 2nd $QS argument */
                        if ((arg4 = intexpr (tmp2)) < -1) ierr = ARGER;
                        if (ierr != OK) break;

                        /* special if source is empty */
                        if (tmp3[0] != EOL || (arg4 != 0)) {
                            /* special: Set env to empty: no |""| */
                            if ((arg4 == -1) && (*argptr == EOL)) {
                                tmp2[0] = EOL;
                            }                        
                            else if ((arg4 != 0) && !znamenumeric (argptr)) {
                                /* put replacement string in tmp2 with     */
                                /* quotes around env or subscript, unless numeric */
                                i = 0;
                                j = -1;
                                tmp2[0] = '"';

                                while ((tmp2[++i] = argptr[++j]) != EOL) {
                                    if (tmp2[i] == '"') tmp2[++i] = '"';

                                    if (i >= (STRLEN - 2)) {
                                        ierr = MXSTR;
                                        break;
                                    }
                                }

                                tmp2[i] = '"';
                                tmp2[++i] = EOL;
                            }
                            else {
                                stcpy (tmp2, argptr);
                            }

                            /* source is tmp3, dest is argptr, replacement is tmp2 */
                            {
                                int ch, cpflag, quote, piececounter;
                                
                                piececounter = 0;
                                i = 0;
                                j = 0;
                                quote = FALSE;
                                cpflag = FALSE;
                                
                                /* if source has no env, process right now */
                                if ((arg4 == -1) && (tmp3[tmp3[0] == '^'] != '|') && tmp2[0] != EOL) {
                                    
                                    if (tmp3[0] == '^') {
                                        argptr[j++] = '^';
                                        i = 1;
                                    }
                                    
                                    argptr[j++] = '|';
                                    ch = 0;
                                    
                                    while ((argptr[j] = tmp2[ch++]) != EOL) j++;

                                    argptr[j++] = '|';

                                }
                                else if (arg4 == 0) {   /* '^'+name may be separated by env */                                
                                    if (tmp2[0] == '^') argptr[j++] = '^';
                                    if (tmp3[0] == '^') i++;
                                }

                                while ((ch = tmp3[i++]) != EOL) {
                                    if (ch == '"') quote = !quote;
                                    
                                    if (!quote) {
                                        
                                        if (ch == ',') {
                                            piececounter++;
                                            argptr[j++] = ch;
                                            
                                            continue;
                                        }
                                        
                                        if ((ch == '(' && piececounter == 0)) {
                                            if (!cpflag && (arg4 == 0)) {
                                                i--;
                                            }
                                            else {
                                                piececounter = 1;
                                                argptr[j++] = ch;

                                                continue;
                                            }
                                        }

                                        if (ch == '|') {
                                            if (piececounter == 0) {
                                                piececounter = (-1);
                                            }
                                            else if (piececounter == (-1)) {
                                                piececounter = 0;
                                            }

                                            if (tmp2[0] != EOL || piececounter > 0) argptr[j++] = ch;

                                            continue;
                                        }
                                    }
                                    
                                    if (piececounter == arg4) {
                                        if (cpflag) continue;

                                        cpflag = TRUE;
                                        ch = 0;
                                        
                                        if (arg4 == 0 && tmp2[0] == '^') ch = 1;

                                        while ((argptr[j] = tmp2[ch++]) != EOL) j++;
                                    }
                                    else {
                                        argptr[j++] = ch;
                                    }

                                    if (j >= (STRLEN - 1)) {
                                        ierr = MXSTR;
                                        break;
                                    }
                                } /* while ((ch = tmp3[i++]) != EOL) ... */

                                if (piececounter && (piececounter == arg4)) argptr[j++] = ')';

                                if (piececounter < arg4) {
                                    
                                    if (piececounter == 0) {
                                        argptr[j++] = '(';
                                    }
                                    else {
                                        argptr[j - 1] = ',';
                                    }

                                    while (++piececounter < arg4) {
                                        argptr[j++] = '"';
                                        argptr[j++] = '"';
                                        argptr[j++] = ',';

                                        if (j >= STRLEN) {
                                            ierr = MXSTR;
                                            break;
                                        }

                                    }
                                }
                                
                                ch = 0;

                                if (argptr[j - 1] != ')') {
                                    while ((argptr[j++] = tmp2[ch++]) != EOL);
                                    argptr[j - 1] = ')';
                                }
                            }

                            argptr[j] = EOL;

                            if (j >= STRLEN) {
                                ierr = MXSTR;
                                break;
                            }
                        
                        }
                    }
                    else {
                        break;
                    }
                } /* set$qsubscript */
                else if (setpiece == 'e') {           /* SETtable $EXTRACT *//* parameters ok?? */

                    if (arg3 > arg4 || arg4 < 1) {
                        setpiece = FALSE;
                        break;
                    }

                    if (arg3 <= 0) arg3 = 1;

                    if (arg3 > STRLEN) {
                        ierr = MXSTR;
                        goto err;
                    }

                    /* get value of glvn */
                    if (vn[0] == '^') { 
                        if (vn[1] == '$') {
                            ssvn (get_sym, vn, tmp3);
                        }
                        else {
                            global (get_sym, vn, tmp3);
                        }
                    }
                    else {
                        symtab (get_sym, vn, tmp3);
                    }


                    /* if UNDEF assume null string */
                    if (ierr == UNDEF || ierr == (UNDEF - CTRLB) || ierr == M6 || ierr == (M6 - CTRLB) || ierr == M7 || ierr == (M7 - CTRLB)) {
                        tmp3[0] = EOL;
                        ierr = OK;
                    }
                    else if (ierr != OK) {
                        stcpy (varerr, vn);
                    }

                    j = stlen (tmp3);

                    /* pad with SPaces if source string is too short */
                    while (j < arg3) tmp3[j++] = SP;

                    tmp3[j] = EOL;
                    
                    if (stlen (tmp3) > arg4) {
                        stcpy (tmp2, &tmp3[arg4]);
                    }
                    else {
                        tmp2[0] = EOL;
                    }

                    tmp3[arg3 - 1] = EOL;

                    /* compose new value of glvn */
                    if (stcat (tmp3, argptr) == 0) {
                        ierr = MXSTR;
                        goto err;
                    }

                    if (stcat (tmp3, tmp2) == 0) {
                        ierr = MXSTR;
                        goto err;
                    }

                    stcpy (argptr, tmp3);
                    setpiece = FALSE;
                }
                else if (setpiece == 'g') {           /* SETtable $GET */
                    setpiece = FALSE;
                    ch = (stcmp (tmp2, argptr) == 0) ? killone : set_sym;

                    if (vn[0] == '^') {
                        if (vn[1] == '$') {
                            ssvn (ch, vn, argptr);
                        }
                        else {
                            global (ch, vn, argptr);
                        }
                    }
                    else {
                        symtab (ch, vn, argptr);
                    }

                    if (ierr != OK) stcpy (varerr, vn);
                    break;
                }

                if (vn[0] == '^') {           /* global variables and SSVNs */
                    
                    if (vn[1] == '$') {
                        ssvn (set_sym, vn, argptr);
                    }
                    else {
                        global (set_sym, vn, argptr);
                    }


                    if (ierr > OK) {
                        stcpy (varerr, vn);
                        goto err;
                    }
                }
                else if (vn[0] != '$') {           /* local variables */
                    symtab (set_sym, vn, argptr);

                    if (ierr > OK) {
                        stcpy (varerr, vn);
                        goto err;
                    }
                }
                else {           /* $-variables */

                    if (vn[1] == 'x') {           /* set $X */
                        j = intexpr (argptr);

                        if (ierr == MXNUM) {
                            j = 256;
                            ierr = OK;
                        }

                        if (j < 0) {
                            ierr = M43;
                            goto err;
                        }

                        if (io == HOME) {
                            argptr[0] = ESC;
                            argptr[1] = '[';
                            argptr[2] = EOL;

                            if (ypos[HOME] > 1) {
                                intstr (tmp3, ypos[HOME] + 1);
                                stcat (argptr, tmp3);
                            }

                            if (j > 0) {
                                stcat (argptr, ";\201");
                                intstr (tmp3, j + 1);
                                stcat (argptr, tmp3);
                            }

                            stcat (argptr, "H\201");
                            write_m (argptr);
                        }

                        xpos[io] = j;
                        goto s_end;
                    }
                    else if (vn[1] == 'y') {           /* set $Y */
                        
                        j = intexpr (argptr);
                        
                        if (ierr == MXNUM)  {
                            j = 256;
                            ierr = OK;
                        }

                        if (j < 0) {
                            ierr = M43;
                            goto err;
                        }

                        if (io == HOME) {

                            argptr[0] = ESC;
                            argptr[1] = '[';
                            argptr[2] = EOL;

                            if (j > 0) {
                                intstr (tmp3, j + 1);
                                stcat (argptr, tmp3);
                            }

                            if (xpos[HOME] > 0) {
                                stcat (argptr, ";\201");
                                intstr (tmp3, xpos[HOME] + 1);
                                stcat (argptr, tmp3);
                            }

                            stcat (argptr, "H\201");
                            write_m (argptr);
                        }
                        
                        ypos[io] = j;
                        goto s_end;
                    }
                    else if (vn[1] == 't') {           /* set $t */
                        test = tvexpr (argptr);
                        goto s_end;
                    }                    
                    else if ((vn[1] == 'r') || ((vn[1] == 'z') && (vn[2] == 'r') && vn[3] == EOL)) { /* set $reference */
                        
                        if (argptr[0] == EOL) {
                            zref[0] = EOL;
                            break;
                        }
                        
                        stcpy (tmp4, codptr);
                        stcpy (code, argptr);
                        
                        codptr = code;
                        
                        expr (NAME);
                        stcpy (code, tmp4);
                        
                        codptr = code;

                        if (argptr[0] != '^') ierr = INVREF;
                        if (ierr <= OK) nakoffs = stcpy (zref, argptr); /* save reference */

                        goto s_end;
                    }
                    else if (vn[1] == 'z') {           /* $Z.. variables *//* if not intrinsic: make it user defined */
                        
                        i = stcpy (&tmp[1], &vn[1]) + 1;
                        
                        if (vn[3] == DELIM) i = 3;  /* set $zf() function keys */

                        tmp[0] = SP;
                        tmp[i] = SP;
                        tmp[++i] = EOL;
                        
                        if (find (zsvn, tmp) == FALSE) {
                            
                            i = 2;
                            while (vn[i] != EOL) {
                                
                                if (vn[i++] == DELIM) {
                                    ierr = INVREF;
                                    goto err;
                                }

                            }

                            udfsvn (set_sym, &vn[2], argptr);
                            break;
                        }
                        
                        if ((!stcmp (&vn[2], "l\201")) || (!stcmp (&vn[2], "local\201"))) { /* set $zlocal */
                            
                            if (argptr[0] == EOL) {
                                zloc[0] = EOL;
                                break;
                            }
                            
                            stcpy (tmp4, codptr);
                            stcpy (code, argptr);
                            
                            codptr = code;
                            
                            expr (NAME);
                            stcpy (code, tmp4);
                            
                            codptr = code;
                            
                            if (argptr[0] == '^') ierr = INVREF;
                            if (ierr <= OK) stcpy (zloc, argptr);   /* save reference */

                            break;
                        }
                        if ((!stcmp (&vn[2], "t\201")) || (!stcmp (&vn[2], "tr\201")) || (!stcmp (&vn[2], "trap\201"))) {       /* set $ztrap */
                            
                            if (stlen (argptr) > ZTLEN) {
                                ierr = MXSTR;
                                goto err;
                            }
                            
                            /* DSM V.2 error trapping */
#ifdef DEBUG_NEWSTACK
                            printf ("Setting Ztrap, DSM2err [%d]\r\n", DSM2err);
#endif


                            if (DSM2err) {
                                stcpy (ztrap[NESTLEVLS + 1], argptr);
                            }
                            else {
                                stcpy (ztrap[nstx], argptr);
                            }

                        }
                        else if (!stcmp (&vn[2], "p\201") || !stcmp (&vn[2], "precision\201")) { /* set $zprecision */

                            short tmp_zprecise;
                            
                            if ((tmp_zprecise = intexpr (argptr)) < 0) {
                                ierr = MXNUM;
                                goto err;
                            }

                            if (!fp_mode) {
                                
                                if (ierr == MXNUM) goto err;

                                if (tmp_zprecise > 20000) {
                                    ierr = MXNUM;
                                    goto err;
                                }
                                
                            }
                            else {
                                
                                if (tmp_zprecise > DBL_DIG) {
                                    ierr = MXNUM;
                                    goto err;
                                }

                                sprintf (fp_conversion, "%%.%df\201", tmp_zprecise);
                                
                            }

                            zprecise = tmp_zprecise;
                            

                        }
                        else if (vn[2] == 'f' && vn[3] == DELIM) {       /* set $zf() function keys */
                            
                            i = intexpr (&vn[4]) - 1;
                            
                            if (i < 0 || i > 43) {
                                ierr = FUNARG;
                                goto err;
                            }

                            if (stlen (argptr) > FUNLEN) {
                                ierr = MXSTR;
                                goto err;
                            }

                            stcpy (zfunkey[i], argptr);

                        }                    
                        else if (vn[2] == 'm' && vn[4] == EOL && (vn[3] == 'c' || vn[3] == 'n' || vn[3] == 'p' || vn[3] == 'l' || vn[3] == 'u')) { /* set $zm_ loadable match; sort match code */
                            
                            short k;

                            i = 0;

                            for (ch = 0; ch <= 255; ch++) {
                                j = argptr - partition;

                                while ((k = partition[j++]) != EOL) {
                                    
                                    if (UNSIGN (k) == ch) {
                                        tmp[i++] = k;
                                        break;
                                    }

                                }

                            }

                            tmp[i] = EOL;
                            
                            switch (vn[3]) {
                            
                                case 'c':
                                    stcpy (zmc, tmp);
                                    break;
                            
                                case 'n':
                                    stcpy (zmn, tmp);
                                    break;

                                case 'p':
                                    stcpy (zmp, tmp);
                                    break;
                                    /*   'a': always union of zml+zmu */

                                case 'l':
                                    stcpy (zml, tmp);
                                    break;
                            
                                case 'u':
                                    stcpy (zmu, tmp);
                                    break;
                                    /*   'e': always 'everything'     */
                            }

                        }
                        else {
                            ierr = INVREF;
                            break;
                        }
                    } 
                    else {
                        ierr = INVREF;
                        goto err;
                    } /* end of processing for $Z.. intrinsic special variables */
                }           /* svns=$vars */

                if (multiset) goto multi;
            } /* end of scope for special SET syntaxes */

s_end:
            if (*codptr != ',') break;
            if (*++codptr == '@') goto again;

            goto set0;

        case IF:

            if (*codptr == SP || *codptr == EOL) {           /* no argument form of IF */
                if (test) goto next_cmnd;
                goto skip_line;
            }

            expr (STRING);

            if (*argptr == '1') {
                test = TRUE;
                break;
            }

            if (ierr > OK) break;
            
            if (*argptr == '0' && argptr[1] == EOL) {
                test = FALSE;
                goto skip_line;
            }

            if ((test = tvexpr (argptr)) == FALSE) goto skip_line;
            break;

        case WRITE:

            {
                
                char *wr_terminator = (char *) malloc (STRLEN * sizeof (char));
                char *sw_buf = (char *) malloc (STRLEN * sizeof (char));


                msck_get_terminator (io, wr_terminator);

                if (io != HOME && devopen[io] == 'r' && io < FIRSTSCK) {
                    ierr = NOWRITE;
                    free (sw_buf);
                    goto err;
                }
                
                if ((i = (*codptr)) == SP || i == EOL) {
                
                    if (standard) {
                        ierr = NOSTAND;
                        free (sw_buf);
                    }
                
                    goto zwrite;
                }

    writeproc:

                switch (i)
                {

                    case '!':

                        if (io < FIRSTSCK) {
                        
                            if (crlf[io]) {
                                write_m ("\012\201");
                            }
                            else {
                                write_m ("\015\012\201");
                            }

                        }
                        else {

                            /* WRITE ! for sockets */
                            strcat (sw_buf, wr_terminator);

                        }

                        break;

                    case '#':

                        if (io < FIRSTSCK) {
                            write_m ("\015\014\201");
                        }
                        else {
                            ierr = SYNTERR;
                            goto err;
                        }

                        break;

                    case '?':

                        codptr++;
                            
                        expr (STRING);

                        if (ierr > OK) goto err;
                        
                        if (io < FIRSTSCK) {
                            write_t ((short) intexpr (argptr));
                        }
                        else {
                            ierr = SYNTERR;
                            goto err;
                        }

                        break;

                    case '/':
                        codptr++;
                        
                        expr (NAME);
                        
                        if (ierr > OK) goto err;
                        
                        if (io < FIRSTSCK) {
                            write_f (varnam);
                        }
                        else {
                            ierr = SYNTERR;
                            goto err;
                        }

                        codptr++;
                        
                        break;

                    case '*':
                        codptr++;
                        
                        expr (STRING);
                        
                        if (ierr > OK) goto err;
                        
                        argptr[0] = (char) UNSIGN (intexpr (argptr));
                        argptr[1] = EOL;
                        
                        /* special treatment for EOL as long as we don't have 8 bit */
                        if (argptr[0] == EOL) {
                            
                            mcmnd = '*';
                            
                            if (io < FIRSTSCK) {
                                m_output (argptr);
                            }
                            else {
                                /* EOL handling for socket devices */
                            }
                            mcmnd = WRITE;

                        }
                        else {

                            if (io < FIRSTSCK) {
                                write_m (argptr);
                            }
                            else {
                                /* non-EOL char-code handling for sockets */
                            }

                        }

                        break;

                    default:
                        expr (STRING);
                        
                        if (ierr > OK) goto err;
                        
                        if (io < FIRSTSCK) {
                            write_m (argptr);
                        }
                        else {
                            /* regular output for socket devices */
                            stcpy (vn, argptr);
                            stcnv_m2c (vn);
                            strcat (sw_buf, vn);
                        }
                }

                if ((i == '!') || (i == '#')) {
                    if ((i = *++codptr) == '!' || i == '#' || i == '?') goto writeproc;
                }

                if (io >= FIRSTSCK) {
                    msck_write (io, sw_buf, strlen (sw_buf));
                }

                free (sw_buf);


                break;
            }

        case READ:

            if (io != HOME && devopen[io] != 'r' && devopen[io] != '+' && io < FIRSTSCK) {
                ierr = NOREAD;
                goto err;
            }

read_command:

            switch (*codptr)
            {
                case '!':
                    if (crlf[io]) {
                        write_m ("\012\201");
                    }
                    else {
                        write_m ("\012\015\201");
                    }
                    
                    if (*++codptr == '!' || *codptr == '#' || *codptr == '?') goto read_command;

                    goto cont_read;

                case '#':
                    write_m ("\015\014\201");

                    if (*++codptr == '!' || *codptr == '#' || *codptr == '?') goto read_command;
                    
                    goto cont_read;
                
                case '?':
                    codptr++;
                    
                    expr (STRING);
                    
                    if (ierr > OK) goto err;
                    
                    write_t ((short) intexpr (argptr));
                    
                    goto cont_read;
                
                case '/':
                    codptr++;
                    
                    expr (NAME);
                    if (ierr > OK) goto err;

                    write_f (varnam);
                    
                    codptr++;
                    
                    goto cont_read;
                
                case '"':
                    i = 0;
                    
                    for (;;) {
                        
                        while ((ch = *++codptr) > '"') argptr[i++] = ch;
                        
                        /* EOL < "any ASCII character" */
                        if (ch == '"' && (ch = *++codptr) != '"') {
                            argptr[i] = EOL;
                            write_m (argptr);

                            goto cont_read;
                        }

                        if ((argptr[i++] = ch) == EOL) {
                            ierr = QUOTER;
                            goto err;
                        }
                    }
            }

            i = InFieldLen;     /* no length limit */
            InFieldLen = 255;       /* Not necessarily tied to STRLEN */

            if (*codptr == '*') {
                codptr++;
                i = 0;
            }           /* single char read */
            
            if (*codptr == '$') {
                ierr = INVREF;
                goto err;
            }

            expr (NAME);
            
            if (ierr > OK) goto err;
            
            stcpy (vn, varnam);
            codptr++;           /* lvn */

            if (i != 0 && *codptr == '#') {           /* length limit */
                codptr++;
            
                expr (STRING);
            
                if ((i = intexpr (argptr)) <= 0) ierr = ARGER;
                if (ierr > OK) goto err;
            }

            timeout = (-1L);
            timeoutms = 0;      /* no timeout */

            if (*codptr == ':')
            {           /* timeout */
                int i, ch;

                codptr++;

                expr (STRING);
                numlit (argptr);
                
                if (ierr > OK) goto err;
                
                timeout = 0;
                timeoutms = 0;
                
                if (argptr[0] != '-') {
                    
                    i = 0;

                    for (;;) { /* get integer and fractional part */ 
                        
                        if ((ch = argptr[i++]) == EOL) break;

                        if (ch == '.') {
                        
                            timeoutms = (argptr[i++] - '0') * 100;
                        
                            if ((ch = argptr[i++]) != EOL) {                        
                                timeoutms += (ch - '0') * 10;
                        
                                if ((ch = argptr[i]) != EOL) {
                                    timeoutms += (ch - '0');
                                }

                            }
                            break;
                        }
                        timeout = timeout * 10 + ch - '0';
                    }
                }
            } /* if (*codptr == ':') */

#if 0
            set_io (UNIX);
            printf ("READ: io = %d timeout = %d timeoutms = %d count = %d\n", io, timeout, timeoutms, i);
            set_io (MUMPS);
#endif            

            if (io < FIRSTSCK) {
                /* $IO is not a socket device */
                read_m (argptr, timeout, timeoutms, i);
            }
            else {
                /* $IO _is_ a socket device */
                msck_read (io, argptr, timeout, timeoutms, i);
            }

            if (vn[0] != '^') {
                symtab (set_sym, vn, argptr);
            }
            else {
                if (vn[1] == '$') {
                    ssvn (set_sym, vn, argptr);
                }
                else {
                    global (set_sym, vn, argptr);
                }
            }

            if (ierr != OK) stcpy (varerr, vn);

cont_read:

            break;

        case ELSE:
            
            if (*codptr != EOL) {
            
                if (*codptr != SP) {
                    ierr = ARGER;
                    break;
                }
            
                if (test == FALSE) goto next_cmnd; /* same line */
            }

            goto skip_line;     /* next line */

        case QUIT:

#ifdef DEBUG_NEWSTACK
            printf ("At QUIT command, checking stack...\r\n");
#endif



#ifdef DEBUG_NEWSTACK
            printf ("nestc[nstx] is (%d)\r\n", nestc[nstx]);
#endif

            if (*codptr != EOL && *codptr != SP && nestc[nstx] != '$') {
#ifdef DEBUG_NEWSTACK
                printf ("IERR\r\n");
#endif

                ierr = ARGER;
                break;
            }



            if (nestc[nstx] == '$') {           /* extrinsic function/variable */


#ifdef DEBUG_NEWSTACK
                printf ("EXTRINSIC\r\n");
#endif

                if (*codptr == EOL || *codptr == SP) {

#ifdef DEBUG_NEWSTACK
                    printf ("CODPTR is [%d]\r\n", *codptr);
#endif

                    if (exfdefault[0] == EOL) {
                        *argptr = EOL;
                        ierr = NOVAL;
                    }
                    else { /* there is a default expression... */
                        stcpy (&code[1], exfdefault);
                        expr (STRING);

                        if (ierr != OK - CTRLB && ierr != OK && ierr != INRPT) {

#ifdef DEBUG_NEWSTACK
                            printf ("Break at 1st IERR\r\n");
#endif
                            break;
                        }
                    }
                }
                else {
                    
                    expr (STRING);

                    if (ierr != OK - CTRLB && ierr != OK && ierr != INRPT) {

#ifdef DEBUG_NEWSTACK
                        printf ("Break at 2nd IERR\r\n");
#endif

                        break;
                    }
                }

#ifdef DEBUG_NEWSTACK
                printf ("CHECK 01 (Stack POP)\r\n");
#endif



                if (nestn[nstx]) {           /* reload routine */
                    namptr = nestn[nstx];

                    stcpy (rou_name, namptr);
                    zload (rou_name);
                    
                    ssvn_job_update ();

                    dosave[0] = 0;
                    
                    namptr--;
                }

                if (nestnew[nstx]) unnew ();       /* un-NEW variables */
                
                /* restore old pointers */
                level = nestlt[nstx];   /* pop level */
                roucur = nestr[nstx] + rouptr;
                
                stcpy (codptr = code, cmdptr = nestp[nstx--]);
                estack--;
                
                forsw = (nestc[nstx] == FOR);
                loadsw = TRUE;
                
                return 0;


            }


            if (nestc[nstx] == BREAK) {
                ierr = OK - CTRLB;
                goto zgo;
            }           /*cont. single step */


quit0:

#ifdef DEBUG_NEWSTACK
            printf ("CHECK 02 (Stack POP)\r\n");
#endif

            if (evt_depth) {
                    
                evt_depth--;
                    
                if (evt_depth == 0 && evt_async_restore == TRUE) {
                    evt_async_enabled = TRUE;
                    evt_async_restore = FALSE;
                }

            }

            if (etrap_lvl) etrap_lvl--;

            if (nstx == 0) goto restore;       /* nothing to quit */

            if (nestc[nstx] == FOR) {
                
                stcpy (code, cmdptr = nestp[nstx--]);

                estack--;
            
                codptr = code;
            
                ftyp = fortyp[--forx];
                fvar = forvar[forx];
                finc = forinc[forx];
                flim = forlim[forx];
                fi = fori[forx];

                if ((forsw = (nestc[nstx] == FOR))) goto for_end;
            
                goto next_line;
            }

            if (nestn[nstx]) {           /* reload routine */
                namptr = nestn[nstx];

                if ((nestc[nstx] != XECUTE) || loadsw) {
                    
                    stcpy (rou_name, namptr);
                    zload (rou_name);
                
                    ssvn_job_update ();

                    dosave[0] = 0;
                }

                namptr--;
            }

            if (nestnew[nstx]) unnew ();       /* un-NEW variables */

            /* restore old pointers */
            if ((mcmnd = nestc[nstx]) == BREAK) goto restore;       /* cont. single step */

            if (mcmnd == DO_BLOCK) {
                test = nestlt[nstx];
                level--;
            }
            else { /* pop $TEST */
                level = nestlt[nstx];   /* pop level */
            }

            if (nstx) {
                roucur = nestr[nstx] + rouptr;
            }
            else {
                roucur = buff + (NO_OF_RBUF * PSIZE0 + 1);
            }

            stcpy (codptr = code, cmdptr = nestp[nstx--]);
            estack--;
            forsw = (nestc[nstx] == FOR);
            
            loadsw = TRUE;

            if (deferred_ierr > OK) {
                ierr = deferred_ierr;
            }

            break;

        case FOR:

            if ((ch = *codptr) == EOL) goto skip_line;     /* ignore empty line */

#ifdef DEBUG_NEWSTACK
            printf ("CHECK 03 (Stack PUSH)\r\n");
#endif


            
            if (++nstx > NESTLEVLS) {
                nstx--;                
                ierr = STKOV;
            
                break;
            }
            else {
                estack++;
            }

            fvar = forvar[++forx];
            finc = forinc[forx];
            flim = forlim[forx];
            fi = fori[forx];
            nestc[nstx] = FOR;  /* stack set-up */

#ifdef DEBUG_NEWSTACK
            if (!cmdptr) printf ("CMDPTR is ZERO!\r\n");
#endif

            nestp[nstx] = cmdptr;
            nestn[nstx] = 0;        /* no overring of routine */
            nestr[nstx] = roucur - rouptr;  /* save roucur: only for $V(26) needed */
            ztrap[nstx][0] = EOL;


            forsw = TRUE;
            ftyp = 0;           /* no args is FOREVER */

            if (ch == SP) {
                goto for_go;
            }
            else {           /* find local variable */
                
                if (ch == '^') {
                    ierr = GLOBER;
                    break;
                }
                
                if (ch == '$') {
                    ierr = INVREF;
                    break;
                }

                if (*(codptr + 1) == '=') {           /* single char local variable */

                    if ((ch < 'A' && ch != '%') || (ch > 'Z' && ch < 'a') || ch > 'z') {
                        ierr = INVREF;
                        break;
                    }

                    fvar[0] = ch;
                    fvar[1] = EOL;
                    
                    codptr += 2;
                }
                else {
                    expr (NAME);

                    if (*++codptr != '=') ierr = ASSIGNER;
                    if (ierr > OK) break;

                    stcpy (fvar, varnam);

                    codptr++;
                }

                ftyp++;
            }

for_nxt_arg:

            expr (STRING);
            
            if (ierr > OK) break;

            stcpy (tmp, argptr);

            if ((ch = *codptr) != ':') {
                
                if (ch == ',' || ch == SP || ch == EOL) {
                    ftyp = 1;
                    goto for_init;
                }
            
                ierr = ARGLIST;
                break;
            }

            numlit (tmp);       /* numeric interpretation */

            codptr++;
            expr (STRING);
            
            if (ierr > OK) break;
            
            numlit (argptr);
            stcpy (finc, argptr);   /* increment */
            
            if ((ch = *codptr) != ':') {
                
                if (ch == ',' || ch == EOL || ch == SP) {
                    ftyp = 2;
                    goto for_init;
                }

                ierr = ARGLIST;
                break;

            }

            codptr++;

            expr (STRING);
            if (ierr > OK) break;

            numlit (argptr);
            stcpy (flim, argptr);   /* limit */

            ftyp = 3;

            if ((ch = *codptr) != ',' && ch != SP && ch != EOL) {
                ierr = ARGLIST;
                break;
            }

            if ((*finc != '-' && comp (flim, tmp)) || (*finc == '-' && comp (tmp, flim))) {
            
                symtab (set_sym, fvar, tmp);
            
                if (ierr > OK) {
                    stcpy (varerr, vn);
                    break;
                }
            
                goto for_quit;
            }

for_init:

            symtab (set_sym, fvar, tmp);

            if (ierr > OK) {
                stcpy (varerr, fvar);
                break;
            }

            /* optimize frequent special case: */
            /* increment by one and no additional FOR arguments */
            /* if limit value it must be a positive integer */
            if (ftyp > 1 && finc[0] == '1' && finc[1] == EOL) {
                j = TRUE;
                
                if (ftyp == 3) {
                    i = 0;

                    while ((ch = flim[i]) != EOL) {
                        
                        if (ch < '0' || ch > '9') j = FALSE;
                    
                        i++;
                    }

                    fi = i;
                    fori[forx] = i;
                }

                if (j && ((ch = *codptr) == SP || ch == EOL)) {
                    ftyp += 2;                
                    if (ch == SP) codptr++;
                }
            }

for_go:

            fortyp[forx] = ftyp;


#ifdef DEBUG_NEWSTACK
            if (!cmdptr) printf ("CMDPTR is ZERO!\r\n");
#endif

            nestp[nstx] = cmdptr;

            cmdptr += stcpy (cmdptr, codptr) + 1;

            if (ftyp > 3) goto next_cmnd;

            /* skip following for arguments if there are any */

for10:

            if (*codptr == SP) goto next_cmnd;

            i = 0;

            while ((((ch = *codptr) != SP) || i) && ch != EOL) {                
                if (ch == '"') i = !i;                

                codptr++;
            }           /* skip rest of FOR list */

            goto next_cmnd;

for_end:            /* end of line return */



#ifdef DEBUG_NEWSTACK
            printf ("For_end: nstx: %d, Nestp: (%d)\r\n", nstx, nestp[nstx]);
#endif

            stcpy (codptr = code, nestp[nstx]); /* restore old pointers */



            switch (ftyp) {
                
                case 5:     /* frequent special case: increment 1 */
                    symtab (getinc, fvar, tmp);

                    /*  compare fvar-value to flim-value */
                    /* fi: i=0; while (flim[i]>='0') i++; */
                    /* Note: EOL<'-'<'.'<'0' tmp has at least one character */
                    ch = '0';
                    j = 1;

                    while (tmp[j] >= ch) j++;
                    
                    if (j < fi) goto next_cmnd;
                    
                    if (j == fi) {
                        j = 0;
                    
                        while (tmp[j] == flim[j]) {
                            if (tmp[j] == EOL) goto next_cmnd;
                            j++;
                        }
                        
                        if (tmp[j] <= flim[j]) goto next_cmnd;
                    }
                    
                    if (flim[0] != '-' && tmp[0] == '-') goto next_cmnd;

                    stcpy (tmp2, "-1\201"); /* correct last inc */
                    add (tmp, tmp2);
                    symtab (set_sym, fvar, tmp);

                    goto for_quit;
                
                case 4:     /* frequent special case: increment 1 without limit */
                    symtab (getinc, fvar, tmp);

                case 0:     /* argumentless FOR */

                    if(argless_forsw_quit == TRUE) {
                        
                        /* if we have a positive QUIT condition, bail from the FOR loop */
                        argless_forsw_quit = FALSE;
                        goto for_quit;
                    }
                    else {

                        /* otherwise, just keep on truckin' */
                        goto next_cmnd;
                    }

                case 3:     /* FOR with increment and limit test */
                    symtab (get_sym, fvar, tmp);
                    numlit (tmp);
                    stcpy (tmp2, finc); /* add may change forinc */
                    add (tmp, tmp2);

                    if (*finc != '-') {
                        if (comp (flim, tmp)) goto for_quit;
                    }
                    else {
                        if (comp (tmp, flim)) goto for_quit;
                    }

                    symtab (set_sym, fvar, tmp);

                    goto for10;

                case 2:     /* FOR with increment without limit test */
                    symtab (get_sym, fvar, tmp);
                    numlit (tmp);
                    stcpy (tmp2, finc); /* add may change forinc */
                    add (tmp, tmp2);

                    symtab (set_sym, fvar, tmp);
                    goto for10;
            }           /* end switch */

for_quit:


            cmdptr = nestp[nstx];


            if (*codptr++ == ',') goto for_nxt_arg;


            nstx--;
            estack--;

            forx--;
            ftyp = fortyp[forx];
            fvar = forvar[forx];
            finc = forinc[forx];
            flim = forlim[forx];
            fi = fori[forx];


            if ((forsw = (nestc[nstx] == FOR))) goto for_end;


            goto next_line;

        case MERGE:

            {
                char lhs[256];
                char rhs[256];

                char k_buf[STRLEN];
                
                expr (NAME);
                if (ierr > OK) break;

                key_to_name (lhs, varnam, 255);
                stcnv_c2m (lhs);

                if (*++codptr != '=') {
                    ierr = ASSIGNER;
                    break;
                }

                codptr++;

                expr (NAME);
                if (ierr > OK) break;

                codptr++;

                key_to_name (rhs, varnam, 255);
                stcnv_c2m (rhs);

                stcpy (k_buf, "ZFRMMLHS\201");
                symtab (set_sym, k_buf, lhs);

                stcpy (k_buf, "ZFRMMRHS\201");
                symtab (set_sym, k_buf, rhs);
                
                stcpy (&tmp3[1], "ZMERGE \201");
                goto private;

                break;

            }

        
        case RLOAD:
            stcpy (&tmp3[1], "zrload \201");
            goto private;


        case RSAVE:
            stcpy (&tmp3[1], "zrsave \201");
            goto private;


        case XECUTE:
            expr (STRING);

            if (ierr > OK) break;
            
            stcpy (tmp, argptr);

            if (*codptr == ':') {           /* argument postcond */
                
                codptr++;
                expr (STRING);
                
                if (ierr > OK) break;
                if (tvexpr (argptr) == FALSE) break;
            }
            
            if (++nstx > NESTLEVLS) {
                nstx--;
                ierr = STKOV;
            
                break;
            }
            else {
                estack++;
            }


#ifdef DEBUG_NEWSTACK
            if (!cmdptr) printf ("CMDPTR is ZERO!\r\n");
#endif

            nestc[nstx] = XECUTE;            
            nestp[nstx] = cmdptr;   /* command stack address */
            nestr[nstx] = roucur - rouptr;  /* save roucur */
            nestlt[nstx] = level;
            
            level = 0;      /* save level */
            nestnew[nstx] = 0;
            ztrap[nstx][0] = EOL;

            while ((*(namptr++)) != EOL);

            stcpy ((nestn[nstx] = namptr), rou_name);   /* save routine name */

            forsw = FALSE;
            loadsw = FALSE;
            cmdptr += stcpy (cmdptr, codptr) + 1;

            stcpy (code, tmp);

            roucur = buff + (NO_OF_RBUF * PSIZE0 + 1);
            codptr = code;

            goto next_cmnd;

        
        case DO:

            if (evt_async_initial == TRUE) {
                evt_async_initial = FALSE;
            }
            else {
                evt_depth++;
            }

            rouoldc = roucur - rouptr;
            namold = 0;

        case GOTO:

do_goto:

            offset = 0;
            label[0] = routine[0] = EOL;
            dofram0 = 0;

            if (((ch = *codptr) != '+') && (ch != '^')) {           /* parse label */  
                          
                if (ch == SP || ch == EOL) {           /* no args: blockstructured DO */

                    if (mcmnd != DO) {
                        ierr = ARGLIST;
                        break;                    
                    }

                    /* direct mode: DO +1 */


                    if (nstx == 0 && roucur >= rouend) {

                        roucu0 = rouptr;
                        goto off1;
                    }

                    mcmnd = DO_BLOCK;
                    roucu0 = roucur;    /* continue with next line */
                    forsw = FALSE;
                
                    goto off2;
                }

                expr (LABEL);
                
                if (ierr > OK) goto err;

                stcpy (label, varnam);

                ch = *++codptr;
            }

            if (ch == '+') {           /* parse offset */
                
                codptr++;
                expr (OFFSET);

                if (ierr > OK) goto err;
                
                offset = intexpr (argptr);
                dosave[0] = EOL;
                
                /* unless argument is numeric, expr returns wrong codptr */
                if ((ch = *codptr) != SP && (ch != EOL) && (ch != ',') && (ch != '^')) ch = *++codptr;
            
            }

            if (ch == '^') {           /* parse routine */
                codptr++;
                expr (LABEL);

                if (ierr > OK) goto err;
                
                stcpy (routine, varnam);
                
                dosave[0] = EOL;
                ch = *++codptr;
                loadsw = TRUE;
            }

            if (ch == '(' && mcmnd == DO) {           /* parse parameter */

                if (offset) {
                    ierr = ARGLIST;
                    goto err;
                }

                if (*++codptr == ')') {
                    ch = *++codptr;
                }
                else {
                    dofram0 = dofrmptr;
                    i = 0;
                    
                    for (;;) {
                        setpiece = TRUE;    /* to avoid error on closing bracket */
                    
                        if (*codptr == '.' && (*(codptr + 1) < '0' || *(codptr + 1) > '9')) {
                            codptr++;
                            
                            expr (NAME);
                            codptr++;
                            
                            *dofrmptr++ = DELIM;    /* to indicate call by name */
                            dofrmptr += stcpy (dofrmptr, varnam) + 1;
                        }
                        else {
                            expr (STRING);
                            dofrmptr += stcpy (dofrmptr, argptr) + 1;
                        }
                    
                        setpiece = FALSE;
                        i++;
                        
                        if (ierr > OK) {
                            dofrmptr = dofram0;
                            goto err;
                        }
                    
                        ch = *codptr++;
                        if (ch == ',') continue;
                    
                        if (ch != ')') {
                            ierr = COMMAER;
                            dofrmptr = dofram0;
                    
                            goto err;
                        }
                    
                        ch = *codptr;
                        break;
                    }
                }
            }

            if (ch == ':') {           /* parse postcond */
                
                codptr++;
                expr (STRING);
                
                if (ierr > OK) {
                    if (dofram0) dofrmptr = dofram0; /* reset frame pointer */
                    goto err;
                }

                if (tvexpr (argptr) == FALSE) {
                    if (*codptr != ',') mcmnd = 0;  /* avoid false LEVEL Error */
                    if (dofram0) dofrmptr = dofram0; /* reset frame pointer */
                
                    break;
                }
            }

            if (mcmnd == GOTO) {           /* GOTO: clear FORs from stack */
                
#ifdef DEBUG_NEWSTACK
                printf ("CHECK 05 Multi-POP on FOR\r\n");
#endif



                while (nestc[nstx] == FOR) {
                
#ifdef DEBUG_NEWSTACK
                    printf ("POP");
#endif

                    cmdptr = nestp[nstx--];
                    estack--;

                    forx--;
                    ftyp = fortyp[forx];
                    fvar = forvar[forx];
                    finc = forinc[forx];
                    flim = forlim[forx];
                    fi = fori[forx];
                }

#ifdef DEBUG_NEWSTACK
                printf ("\r\n");
#endif


                loadsw = TRUE;
            }

job_entry:     /* entry called from successful JOB */

            if (routine[0] != EOL) {

#ifdef DEBUG_NEWSTACK
                printf ("CHECK 06\r\n");
#endif

                if (mcmnd == DO) {

                    while ((*(namptr++)) != EOL);
                    
                    namold = namptr;
                    stcpy (namptr, rou_name);

                    ssvn_job_update ();

                }

                    /* if (GOTO label^rou) under a (DO label)   */
                    /* save away old routine to restore on quit */


                
                else if (nstx > 0) {

#ifdef DEBUG_NEWSTACK
                    printf ("CHECK 06, stack is greater than 0\r\n");
#endif

                    while (nestc[nstx] == FOR) {
#ifdef DEBUG_NEWSTACK
                        printf ("POP");
#endif

                        nstx--;
                        estack--;
                        forx--;
                        ftyp = fortyp[forx];
                        fvar = forvar[forx];
                        finc = forinc[forx];
                        flim = forlim[forx];
                        fi = fori[forx];
                    }

                    if (nestn[nstx] == 0) {                        
                        while ((*(namptr++)) != EOL);

                        stcpy ((nestn[nstx] = namptr), rou_name);
                    }
                }

                zload (routine);                
                if (ierr > OK) goto err;       /* load file */

                ssvn_job_update ();
            
            } /* if (routine[0] != EOL) */
            {
                char *reg, *reg1;

                reg1 = rouptr;
                reg = reg1;

                if (label[0] != EOL) {
                
                    if (forsw && mcmnd == DO && stcmp (label, dosave) == 0) {
                        roucu0 = xdosave;
                        goto off1;
                    }

                    while (reg < rouend) {
                        reg++;
                        j = 0;
                    
                        while (*reg == label[j]) {
                            reg++;
                            j++;
                        }

                        if (label[j] == EOL) {
                            
                            if (*reg == TAB || *reg == SP) goto off;
                            
                            /* call of procedure without specifying a parameter list */
                            if (*reg == '(') {                                
                                if (dofram0 == 0) dofram0 = dofrmptr;                            
                                goto off;
                            }

                        }

                        reg = (reg1 = reg1 + UNSIGN (*reg1) + 2);
                    }
                    {
                        ierr = M13;
                        stcpy (varerr, label);  /* to be included in error message */
                    
                        if (dofram0) dofrmptr = dofram0; /* reset frame pointer */
                    
                        zload (rou_name);
                        ssvn_job_update ();                        
                        goto err;
                    }
                }
off:
                if (label[0] == EOL && offset > 0) offset--;
                while (offset-- > 0) reg1 = reg1 + (UNSIGN (*reg1) + 2);
                
                if (forsw) {
                    xdosave = reg1;
                    stcpy (dosave, label);
                }

                roucu0 = reg1;
            }

            if (roucu0 >= rouend) {
                ierr = M13;
                stcpy (varerr, label);  /* to be included in error message */

                if (dofram0) dofrmptr = dofram0; /* reset frame pointer */
                
                zload (rou_name);
                ssvn_job_update ();
                goto err;
            }

off1:
            if (routine[0] != EOL) stcpy (rou_name, routine);

            ssvn_job_update ();

            roucu0++;
            forsw = FALSE;

            if (mcmnd != DO) {           /* i.e. GOTO or JOB */
                roucur = roucu0;
                goto off3;
            }

off2:

#ifdef DEBUG_NEWSTACK
            printf ("CHECK 07 (Stack PUSH)\r\n");
#endif



            if (++nstx > NESTLEVLS) {
                nstx--;
                ierr = STKOV;
            
                goto err;
            }
            else {
                on_frame_entry ();
                estack++;
            }

            nestc[nstx] = mcmnd;

#ifdef DEBUG_NEWSTACK
            if (!cmdptr) printf ("CMDPTR is ZERO!\r\n");
#endif

            nestp[nstx] = cmdptr;
            nestn[nstx] = namold;
            nestr[nstx] = rouoldc;
            nestnew[nstx] = 0;

            if (mcmnd != DO_BLOCK) {
                nestlt[nstx] = level;
                level = 0;
            }
            else { /* push level ; clr level */
                nestlt[nstx] = test;
                level++;
            }           /* push $TEST ; inc level */

            ztrap[nstx][0] = EOL;


            cmdptr += stcpy (cmdptr, codptr) + 1;
            roucur = roucu0;

            /* processing for private Z-Command: */
            if (privflag) {



#ifdef DEBUG_NEWPTR
                printf ("Xecline 01 (using NEWPTR): ");
                printf ("[nstx] is [%d], [nestnew] is [%d]", nstx, nestnew[nstx]);
                printf ("- Initialized to newptr\r\n");
#endif /* Debug */

                nestnew[nstx] = newptr;


                stcpy (vn, zargdefname);

                /*was:      vn[0] = '%';   vn[1] = EOL; */

                symtab (new_sym, vn, "");
                /*djw change 'input variable for Z command' to get value of $V(202) */
                /*was:      vn[0] = '%';   vn[1] = EOL; */

                stcpy (vn, zargdefname);
                symtab (set_sym, vn, tmp2);

                privflag = FALSE;
            }

off3:

            if (dofram0) {
                char *reg, *reg1;

                reg = roucu0;
                reg1 = dofram0;
            
                while ((ch = (*reg++)) != '(') {
                    if (ch == SP || ch == TAB || ch == EOL) break;
                }

                if (ch != '(') {
                    ierr = TOOPARA;
                    dofrmptr = dofram0;
            
                    goto err;
                }
            
                j = 0;

                while ((ch = (*reg++)) != EOL) {
            
                    if ((ch == ',' && j) || ch == ')') {
                        varnam[j] = EOL;



#ifdef DEBUG_NEWPTR
                        printf ("Xecline 02: ");
                        printf ("[nstx] is [%d], [nestnew] is [%d]\r\n", nstx, nestnew[nstx]);
#endif

                        if (nestnew[nstx] == 0) nestnew[nstx] = newptr;
            


                        if (reg1 < dofrmptr) {

                            if (*reg1 == DELIM) {   /* call by reference */
                            
                                if (stcmp (reg1 + 1, varnam)) {   /* are they different?? */
                                    symtab (new_sym, varnam, "");
                                    symtab (m_alias, varnam, reg1 + 1);
                                }

                            }
                            else {
                                symtab (new_sym, varnam, "");   /* call by value */
                                symtab (set_sym, varnam, reg1);
                            }

                            reg1 += stlen (reg1) + 1;
                        }
                        else {
                            symtab (new_sym, varnam, "");
                        }

                        if (ch == ')') break;
                
                        j = 0;
                        continue;
                    }

                    if ((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z') || (ch >= '0' && ch <= '9' && j) || (ch == '%' && j == 0)) {
                        varnam[j++] = ch;
                        continue;
                    }

                    ierr = ARGLIST;
                    dofrmptr = dofram0; /* reset frame pointer */
                    
                    goto err;
                }

                if (reg1 < dofrmptr) {
                    ierr = TOOPARA;
                    dofrmptr = dofram0;

                    goto err;
                }

                dofrmptr = dofram0;
            }

            goto next_line;

        /* ZJOB *//* same as JOB, but without timeout */
        /* not recommended; just for backward compatibility */
        case ZJOB:
            if (standard) {
                ierr = NOSTAND;
                goto err;
            }

        case JOB:

            if (*codptr == SP || *codptr == EOL) {
                ierr = M13;
                varerr[0] = EOL;    /* to be included in error message */

                break;
            }

            loadsw = TRUE;
            offset = 0;
            timeout = (-1L);
            label[0] = routine[0] = EOL;

            if (((ch = *codptr) != '+') && (ch != '^')) {           /* parse label */
                expr (LABEL);
                if (ierr > OK) goto err;

                stcpy (label, varnam);

                ch = *++codptr;
            }

            if (ch == '+') {           /* parse offset */
                codptr++;

                expr (OFFSET);
                if (ierr > OK) goto err;

                offset = intexpr (argptr);
                
                /* unless argument is numeric, expr returns wrong codptr */
                if ((ch = *codptr) != SP && (ch != EOL) && (ch != ',') && (ch != '^')) ch = *++codptr;
            
            }

            if (ch == '^') {           /* parse routine */
                codptr++;
                
                expr (LABEL);
                if (ierr > OK) goto err;

                stcpy (routine, varnam);

                dosave[0] = EOL;
                ch = *++codptr;
            }

            dofram0 = NULL;
            if (ch == '(') {           /* parse parameter */
                
                if (offset) {
                    ierr = ARGLIST;
                    goto err;
                }
                
                codptr++;
                dofram0 = dofrmptr;
                
                i = 0;
                for (;;) {
                    setpiece = TRUE;    /* to avoid error on closing bracket */
                    
                    if (*codptr == '.' && (*(codptr + 1) < '0' || *(codptr + 1) > '9')) {
                        codptr++;
                        
                        expr (NAME);
                        
                        codptr++;
                        
                        *dofrmptr++ = DELIM;    /* to indicate call by name */
                        dofrmptr += stcpy (dofrmptr, varnam) + 1;
                    }
                    else {
                        expr (STRING);
                        dofrmptr += stcpy (dofrmptr, argptr) + 1;
                    }

                    setpiece = FALSE;
                    i++;
                    
                    if (ierr > OK) {
                        dofrmptr = dofram0;
                        goto err;
                    }

                    ch = *codptr++;
                    if (ch == ',') continue;

                    if (ch != ')') {
                        ierr = COMMAER;
                        dofrmptr = dofram0;

                        goto err;
                    }

                    ch = *codptr;
                    break;
                }
            }

            if (ch == ':' && *(codptr + 1) == ch) {
                codptr++;       /* timeout,no jobparams */                
            }
            else if (ch == ':' && *(codptr + 1) == '(') { /* parse any 'job parameters', but ignore them otherwise */
                codptr++;
                setpiece = TRUE;    /* to avoid bracket error at end of jobparameters */

                for (;;) {
                    if (*++codptr != ':') expr (STRING);
                    if (*codptr == ':') continue;
                    if (*codptr++ != ')') ierr = ARGER;

                    break;
                }

                setpiece = FALSE;
                ch = (*codptr);
            }

            if (ch == ':') {           /* timeout */
                codptr++;
                expr (STRING);

                if ((timeout = intexpr (argptr)) < 0L) timeout = 0L;
                if (ierr > OK) goto err;

                test = TRUE;
            }

            if (mcmnd == ZJOB) timeout = 0L;       /* ZJOB-command has timeout 0 */

            close_all_globals ();   /* close all globals */
            j = getpid ();      /* job number of father process */

            if (lonelyflag) {           /* single user */
                if (timeout < 0L) {
                    ierr = PROTECT; /* error without timeout */
                }
                else {
                    test = FALSE;   /* timeout always fails */
                }
                
                break;
            }

            while ((i = fork ()) == -1) {
            
                if (timeout == 0L) {
                    test = FALSE;
                    break;
                }
                
                if (timeout > 0L) timeout--;
                
                sleep (1);

            }

            if (mcmnd == ZJOB && zjobflag) {
                
                if (i == 0) {           /* we are in child process */
                    intstr (zb, j); /* $JOB of father job */
                    
                    father = j;
                    pid = getpid ();    /* this is our new job number */

                    jobtime = time (0L);;


                    nstx = 0;       /* clear stack */
                    estack = 0;

                    forx = 0;
                    forsw = FALSE;
                    level = 0;
                    cmdptr = cmdstack;  /*  -  command stack pointer */
                    namptr = namstck;   /*  -  routine name stack pointer */
                    usermode = 0;   /* application mode */
                    ierr = OK;

                    lock (" \201", -1, 'j');    /* tell lock about JOB */
                    goto job_entry;
                }

                /* ignore signal while here */
                sig_attach (SIGUSR1, SIG_IGN);

                while (wait (&zsystem) != i);

                sig_attach (SIGUSR1, &oncld);   /* restore handler */

                ierr = OK;      /* there might be a INRPT from other job */
    
                set_io (MUMPS);
                break;
            }

            if (i == 0) {           /* we are in child process */
                
                intstr (zb, j); /* $JOB of father job */
                father = j;

                pid = getpid ();    /* $J = process ID */
                usermode = 0;       /* no programmer mode */
                DSW |= BIT0;        /* disable echo */
                zbreakon = DISABLE; /* disable CTRL/B */
                breakon = DISABLE;  /* disable CTRL/C */
                hardcopy = DISABLE; /* disable hardcopy function */
                
                fclose (stdin); /* close normal input */
                
                jour_flag = 0;  /* no protocol */
                

                nstx = 0;       /* clear stack */
                estack = 0;

                forx = 0;
                forsw = FALSE;
                level = 0;
                
                cmdptr = cmdstack;  /*  -  command stack pointer */
                namptr = namstck;   /*  -  routine name stack pointer */
                
                /* init random number */
                if ((nrandom = time (0L) * getpid ()) < 0) nrandom = (-nrandom);
                
                ierr = OK;
                lock (" \201", -1, 'j');    /* tell lock about JOB */

                goto job_entry;
            }

            intstr (zb, i);     /* $JOB of the process just started */
            break;

        case KILL:

            /* argumentless: KILL all local variables */
            if (((ch = *codptr) == SP) || ch == EOL) {
                symtab (kill_all, "", "");
                break;
            }

            if (ch != '(') {
                expr (NAME);
            
                /* aviod a disaster if someone types KILL ^PATDAT[TEST] ! */
                if (((ch = *++codptr) != SP) && ch != EOL && ch != ',') ierr = INVREF;
                if (ierr > OK) goto err;
            
                if (varnam[0] == '^') {
                    if (varnam[1] != '$') {
                        global (kill_sym, varnam, tmp);
                    }
                    else {
                        ssvn (kill_sym, varnam, tmp);
                    }
                    break;
                }

                symtab (kill_sym, varnam, tmp);
                break;
            }

            /* exclusive kill */
            tmp[0] = SP;
            tmp[1] = EOL;

            for (;;) {

                codptr++;
                expr (NAME);
                
                if (ierr > OK) goto err;
                
                if (varnam[0] == '^') {
                    ierr = GLOBER;
                    goto err;
                }
                
                i = 0;
                while (varnam[i] != EOL) {

                    if (varnam[i] == DELIM) {
                        ierr = SBSCR;
                        goto err;
                    }

                    i++;
                }

                if (stcat (tmp, varnam) == 0) {
                    ierr = MXSTR;
                    goto err;
                }

                if (stcat (tmp, " \201") == 0) {
                    ierr = MXSTR;
                    goto err;
                }

                if ((ch = *++codptr) == ')') {
                    codptr++;
                    break;
                }

                if (ch != ',') {
                    ierr = COMMAER;
                    goto err;
                }
            }

            symtab (killexcl, tmp, "");
            break;

        case NEWCMD:
        /*case ZNEW:*/

            /* argumentless: NEW all local variables */
            if (((ch = *codptr) == SP) || ch == EOL) {
                ch = nstx;

                while (nestc[ch] == FOR) ch--;       /* FOR does not define a NEW level */
                
#ifdef DEBUG_NEWPTR
                printf ("Xecline 03: (TODO - NEW ALL) ");
                printf ("[ch] is %d, [nestnew] is %d\r\n", ch, nestnew[ch]);
#endif

                if (nestnew[ch] == 0) nestnew[ch] = newptr;
                
                symtab (new_all, "", "");
                break;
            }

            if (ch != '(') {
                expr (NAME);
                
                if (ierr > OK) goto err;

                codptr++;
                
                if (varnam[0] == '^') {
                    ierr = GLOBER;
                    goto err;
                }

                if (varnam[0] == '$') {
                    i = 0;
                
                    while ((ch = varnam[++i]) != EOL) if (ch >= 'A' && ch <= 'Z') varnam[i] = ch + 32;

                    /* set $reference */    
                    if ((stcmp (&varnam[1], "r\201")) && (stcmp (&varnam[1], "reference\201")) && (stcmp (&varnam[1], "zr\201")) && (stcmp (&varnam[1], "zreference\201")) &&
                        (stcmp (&varnam[1], "t\201")) && (stcmp (&varnam[1], "test\201")) && (stcmp (&varnam[1], "j\201")) && (stcmp (&varnam[1], "job\201")) &&
                        (stcmp (&varnam[1], "zi\201")) && (stcmp (&varnam[1], "zinrpt\201")) && (stcmp (&varnam[1], "et\201")) && (stcmp (&varnam[1], "etrap\201")) &&
                        (stcmp (&varnam[1], "es\201")) && (stcmp (&varnam[1], "estack\201"))) {
                            ierr = INVREF;
                            goto err;
                    }
                }


                
                ch = nstx;
                
                while (nestc[ch] == FOR) ch--;       /* FOR does not define a NEW level */

#ifdef DEBUG_NEWPTR
                printf ("Xecline 04 (DANGER): ");
                printf ("[ch] is %d, [nestnew] is %d\r\n", ch, nestnew[ch]);
#endif

                if (nestnew[ch] == 0) nestnew[ch] = newptr;
                
                symtab (new_sym, varnam, "");
                break;
            }

            /* exclusive new */
            tmp[0] = SP;
            tmp[1] = EOL;
            
            for (;;) {
                codptr++;
                expr (NAME);

                if (ierr > OK) goto err;

                if (varnam[0] == '^') {
                    ierr = GLOBER;
                    goto err;
                }

                if (varnam[0] == '$') {
                    ierr = INVREF;
                    goto err;
                }

                i = 0;
                while (varnam[i] != EOL) {

                    if (varnam[i] == DELIM) {
                        ierr = SBSCR;
                        goto err;
                    }

                    i++;
                }

                if (stcat (tmp, varnam) == 0) {
                    ierr = MXSTR;
                    goto err;
                }

                if (stcat (tmp, " \201") == 0) {
                    ierr = MXSTR;
                    goto err;
                }

                if ((ch = *++codptr) == ')') {
                    codptr++;
                    break;
                }

                if (ch != ',') {
                    ierr = COMMAER;
                    goto err;
                }
            }

            ch = nstx;
            while (nestc[ch] == FOR) ch--;           /* FOR does not define a NEW level */

#ifdef DEBUG_NEWPTR
            printf ("Xecline 05 (TODO): ");
            printf ("[ch] is %d, [nestnew] is %d\r\n", ch, nestnew[ch]);
#endif

            if (nestnew[ch] == 0) nestnew[ch] = newptr;

            symtab (newexcl, tmp, "");
            break;

        case LOCK:

            /* argumentless: UNLOCK */
            if ((ch = *codptr) == SP || ch == EOL) {
                lock (" \201", -1, LOCK);
                break;
            }

            if (ch == '+' || ch == '-') {
                tmp[0] = ch;
                ch = (*++codptr);
            }
            else {
                tmp[0] = SP;
            }

            if (ch != '(') {
                expr (NAME);
            
                if (ierr > OK) goto err;

                stcpy (&tmp[1], varnam);
                stcat (tmp, "\001\201");
            }
            else {           /* multiple lock */
                tmp[1] = EOL;

                for (;;) {
                    codptr++;
                    expr (NAME);
                
                    if (ierr > OK) goto err;
                    
                    stcat (tmp, varnam);
                    stcat (tmp, "\001\201");
                    
                    if ((ch = *++codptr) == ')') break;
                    
                    if (ch != ',') {
                        ierr = COMMAER;
                        goto err;
                    }
                }
            }

            timeout = (-1L);        /* no timeout */
            
            if (*++codptr == ':') {
                codptr++;
                expr (STRING);
            
                timeout = intexpr (argptr);
            
                if (ierr > OK) goto err;
                if (timeout < 0L) timeout = 0L;
            }

            lock (tmp, timeout, LOCK);
            break;

        case USE:

            if (*codptr == SP || *codptr == EOL) {
                ierr = ARGER;
                goto err;
            }

            expr (STRING);
            j = intexpr (argptr);

            if (j > MAXSEQ && j < MAXDEV) {
                io = j;
                goto use_socket;
            }

            if (j < 0 || j > MAXDEV) {
                ierr = NODEVICE;
            }
            else if (j != HOME && devopen[j] == 0) {
                ierr = NOPEN;
            }

            if (ierr > OK) goto err;

            io = j;

            if (io == HOME && *codptr == ':' && *(codptr + 1) == '(') {

use0:          /* entry point for processing of device parameters */

                codptr += 2;
                j = 1;
                setpiece = TRUE;    /* so a surplus closing bracket will not be an error */
                
                while (*codptr != ')') {

                    if (*codptr == ':') {
                        codptr++;
                        j++;
                    
                        continue;
                    }
                    
                    expr (STRING);
                    
                    if (ierr > OK) {
                        setpiece = FALSE;
                        goto err;
                    }

                    switch (j) {

                        case 1:
                            i = intexpr (argptr);

                            if (i < 0) i = 0;
                            if (i > 255) i = 255;
                            
                            RightMargin = i;                            
                            break;
                        
                        case 3:
                            i = intexpr (argptr);
                            
                            if (i < 0) i = 0;
                            if (i > 255) i = 255;

                            InFieldLen = i;
                            break;
                        
                        case 5:
                            DSW = intexpr (argptr);
                            break;
                        
                        case 7:
                            i = intexpr (argptr);
                            ypos[HOME] = i / 256;
                            xpos[HOME] = i % 256;
                            
                            if (DSW & BIT7) {
                                
                                i = io;
                                io = HOME;
                                argptr[0] = ESC;
                                argptr[1] = '[';
                                argptr[2] = EOL;

                                if (ypos[HOME]) {
                                    intstr (&argptr[2], ypos[HOME] + 1);
                                }

                                if (xpos[HOME]) {
                                    tmp3[0] = ';';
                                    
                                    intstr (&tmp3[1], xpos[HOME] + 1);
                                    stcat (argptr, tmp3);
                                }

                                stcat (argptr, "H\201");
                                write_m (argptr);
                                
                                io = i;
                            }
                            break;
                        
                        case 9:
                            i = 0;
                            j = 0;

                            while ((ch = argptr[i++]) != EOL) LineTerm[j++] = ch;
                            
                            LineTerm[j] = EOL;
                            break;
                        
                        case 10:
                            BrkKey = (*argptr);
                            
                            /* make new break active */
                            set_io (UNIX);
                            set_io (MUMPS);
                    }
                }
                
                setpiece = FALSE;
                codptr++;
                
                break;
            }
            else if (*codptr == ':') {
                codptr++;
                
                if (io == HOME) {           /* old syntax: enable/disable echo */
                    expr (STRING);
                
                    if (ierr > OK) goto err;
                    
                    if (tvexpr (argptr)) {
                        DSW &= ~BIT0;
                    }
                    else {
                        DSW |= BIT0;
                    }

                }
                else {

                    if (*codptr == '(') {
                        codptr++;
                        setpiece = TRUE;
                    }

                    j = 1;
                    
                    while (*codptr != ')') {

                        if (*codptr == ':') {
                            codptr++;
                            j++;
                            
                            continue;
                        }
                        else if (setpiece == FALSE) {
                            ierr = SPACER;
                            goto err;
                        }

                        expr (STRING);
                    
                        if (ierr > OK) {
                            setpiece = FALSE;
                            goto err;
                        }
                    
                        switch (j) {
                            
                            case 1:
                                fseek (opnfile[io], (long) intexpr (argptr), 0);
                                break;

                            case 2:
                                crlf[io] = tvexpr (argptr);
                                break;
                            
                            case 3:
                                fm_nodelay[io] = tvexpr (argptr);
                                break;
                        }
                    
                        if (setpiece == FALSE) break;
                    }
                    
                    if (setpiece) {
                        codptr++;
                        setpiece = FALSE;
                    }

                    break;
                }
            }
            break;


use_socket:
            {
                char use_parm[256];
                int upct = 0;                

                if (*codptr == ':') {
                    codptr++;
                }
                else {
                    while ((ch = *(codptr++)) != SP && ch != EOL);
                    codptr--;
                    break;
                }

                if (*codptr != '/') {
                    ierr = ARGLIST;
                    goto err;
                }

                codptr++;

                while ((ch = *codptr++) != SP && ch != EOL && isalpha (ch)) {
                    use_parm[upct++] = ch;
                }

                use_parm[upct] = NUL;

                for (upct = 0; upct < strlen (use_parm); upct++) {
                    use_parm[upct] = toupper (use_parm[upct]);
                }

                if (strcmp (use_parm, "CONNECT") == 0) {

                    msck_connect (io);

                    if (ierr > OK) goto err;

                }
                else if (strcmp (use_parm, "BIND") == 0) {
                    write_m("BIND\r\n\201");
                }
                else {
                    ierr = ARGLIST;
                    goto err;
                }

                break;

            }

        case OPEN:

            {
                short k;

                if (*codptr == SP || *codptr == EOL) {
                    ierr = FILERR;
                    goto err;
                }

                expr (STRING);
                k = intexpr (argptr);
                
                if (ierr > OK) goto err;

                
                if (k < 0 || k > MAXDEV) {
                    ierr = NODEVICE;
                    goto err;
                }

                if (k > MAXSEQ) goto open_socket;

                crlf[k] = FALSE;
                fm_nodelay[k] = FALSE;
                xpos[k] = 0;
                ypos[k] = 0;

                /* OPEN implies a previous CLOSE on same channel */
                if ((k != HOME) && devopen[k]) {
                    
                    fclose (opnfile[k]);
                    devopen[k] = 0;
                
                    if (io == k) io = HOME;
                }

                /* process device parameters on HOME at USE command. */
                if (k == HOME && *codptr == ':' && *(codptr + 1) == '(') goto use0;
                
                if (*codptr != ':') {

                    if (k == HOME) break;
                    
                    if (dev[k][0] == EOL) {
                        ierr = FILERR;
                        goto err;
                    }

                    goto open10;
                }

                codptr++;
                
                if (k == HOME) {

                    if (*codptr != ':') {           /* turn echo on/off */
                        
                        expr (STRING);
                        
                        if (ierr > OK) goto err;

                        if (tvexpr (argptr)) {
                            DSW &= ~BIT0;
                        }
                        else {
                            DSW |= BIT0;
                        }
                    }

                    if (*codptr == ':') {           /* dummy timeout on HOME */
                        codptr++;
                        expr (STRING);
                    
                        if (ierr > OK) goto err;
                        
                        test = TRUE;
                    }
                }
                else {
                    int op_pos;

                    expr (STRING);

                    if (ierr > OK) goto err;
                    
                    stcpy (dev[k], argptr);
                    timeout = (-1L);
                    
                    if (*codptr == ':') {
                        
                        codptr++;

                        expr (STRING);                        
                        timeout = intexpr (argptr);
                        
                        if (ierr > OK) goto err;
                        if (timeout < 0L) timeout = 0L;
                    }

open10:

                    j = stcpy (tmp, dev[k]);
                    i = dev[k][j - 1];
                    
                    while (--j >= 0) {
                        if (dev[k][j] == '/') break;
                    }

                    stcpy (tmp2, dev[k]);
                    
                    if (j <= 0) {
                        tmp2[stlen (tmp2)] = NUL;
                        tmp[1] = 'r';
                        i = '+';
                    }                  
                    else { /* default is read+write */
                        tmp2[j] = NUL;

                        j = stcpy (&tmp[1], &tmp[j + 1]);
                        
                        tmp[0] = SP;
                        tmp[j + 1] = SP;
                        tmp[j + 2] = EOL;
                        
                        j = 0;
                        
                        while ((ch = tmp[++j]) != EOL) if (ch >= 'A' && ch <= 'Z') tmp[j] = ch + 32;
                        
                        if (find (" r w a r+ w+ a+ read write append read+ write+ append+ \201", tmp) == FALSE) {
                            tmp[1] = 'r';
                            i = '+';
                        
                            tmp2[strlen (tmp2)] = '/';
                        }
                    }

                    tmp[0] = tmp[1];
                    tmp[1] = NUL;   /* NUL not EOL !!! */
                    
                    if (i == '+') {
                        tmp[1] = i;
                        tmp[2] = NUL;
                    }

                    op_pos = 0;
                    open20:;
                    
                    if (oucpath[op_pos] != EOL) {

                        j = stlen (dev[k]);
                        
                        while (--j >= 0) if (dev[k][j] == '/') break;                        
                        while (--j >= 0) if (dev[k][j] == '/') break;
                        
                        if (j < 0) {

                            strcpy (tmp3, tmp2);
                            stcpy (tmp2, &oucpath[op_pos]);
                            
                            j = 0;
                            while (tmp2[j] != ':' && tmp2[j] != EOL) j++;

                            tmp2[j] = EOL;
                            
                            stcpy (act_oucpath[k], tmp2);
                            op_pos += j;
                            
                            if (j) tmp2[j++] = '/';

                            strcpy (&tmp2[j], tmp3);
                        }
                    }

                    /* r  = READ only access;
                    * w  = WRITE new file;
                    * a  = WRITE append;
                    * r+ = READ/WRITE access;
                    * w+ = WRITE new file;
                    * a+ = WRITE append;
                    */
                    j = tmp[0];
                    if (j == 'r' && timeout < 0L) {

                        errno = 0;
                        
                        while ((opnfile[k] = fopen (tmp2, tmp)) == NULL) {
                            
                            if (errno == EINTR) {
                                errno = 0;
                                continue;
                            }       /* interrupt */
                            
                            if (errno == EMFILE || errno == ENFILE) {
                                close_all_globals ();
                                continue;
                            }

                            if (dev[k][0] != '/' && oucpath[op_pos++] != EOL) {
                                strcpy (tmp2, tmp3);
                                goto open20;
                            }

                            act_oucpath[k][0] = EOL;
                            ierr = (errno == ENOENT ? FILERR : PROTECT);
                            
                            goto err;
                        }

                        devopen[k] = ((i == '+') ? i : j);
                        break;
                    }

                    if (j == 'r' || j == 'w' || j == 'a') {
                        
                        if (timeout >= 0L) {
                            
                            test = TRUE;
                            
                            if (setjmp (sjbuf)) {
                                test = FALSE;
                                goto endopn;
                            }
                            
                            sig_attach (SIGALRM, &ontimo);
                            alarm ((unsigned) (timeout < 3 ? 3 : timeout));
                        }

                        for (;;) {
                            errno = 0;

                            if ((opnfile[k] = fopen (tmp2, tmp)) != NULL) break;
                            if (ierr == INRPT) goto err;
                            if (errno == EINTR) continue;   /* interrupt */
                            
                            if (errno == EMFILE || errno == ENFILE) {
                                close_all_globals ();
                                continue;
                            }

                            if (timeout < 0L) {

                                if (dev[k][0] != '/' && oucpath[op_pos++] != EOL) {
                                    strcpy (tmp2, tmp3);
                                    goto open20;
                                }

                                if (errno == ENOENT) continue;

                                act_oucpath[k][0] = EOL;
                                ierr = PROTECT;
                                
                                goto err;
                            }

                            if (timeout == 0L) {
                                test = FALSE;
                                goto endopn;
                            }

                            sleep (1);
                            timeout--;
                        }

                        devopen[k] = ((i == '+') ? i : j);
                        endopn:;
                        
                        alarm (0);  /* reset alarm request */
                    }
                    else {
                        ierr = ARGLIST;
                        goto err;
                    }
                }


open_socket:
            
                if (*codptr != ':') {
                    ierr = ARGLIST;
                    goto err;
                }

                codptr++;
                expr (STRING);

                if (ierr > OK) goto err;


                stcpy (vn, argptr);
                stcnv_m2c (vn);

                msck_open (k, vn);

                break; 
            }
            break;

          



        case CLOSE:

            /* no arguments: close all exept HOME */
            if (*codptr == SP || *codptr == EOL) {

                if (standard) {
                    ierr = NOSTAND;
                    break;
                }

                j = 1;
                
                while (j <= MAXDEV) {
                    
                    if (j < FIRSTSCK) {
                        if (jour_flag && (j == 2)) {
                            j++;
                            continue;
                        }
                
                        if (devopen[j]) fclose (opnfile[j]);
                
                        devopen[j++] = 0;
                    }
                    else {
                        msck_close (j++);
                    }

                }

                io = HOME;
                break;
            }

            expr (STRING);
            j = intexpr (argptr);

            if (ierr > OK) break;

            if (j >= FIRSTSCK && j < MAXDEV) {
                msck_close (j);
                break;
            }

            /*ignore close on illgal units */
            if ((j >= 0 && j <= MAXDEV && j != HOME) && (jour_flag == 0 || (j != 2))) {           /*ignore close on protocol channel */
                
                if (devopen[j]) fclose (opnfile[j]);
                
                devopen[j] = 0;
                
                if (io == j) io = HOME;

            }

            /* parse any 'device parameters', but ignore them otherwise */
            if (*codptr == ':') {
                if (*++codptr != '(') {
                    expr (STRING);
                }
                else {
                    setpiece = TRUE;    /* to avoid bracket error at end of deviceparameters */
                    for (;;)
                    {
                    if (*++codptr != ':')
                    expr (STRING);
                    if (*codptr == ':')
                    continue;
                    if (*codptr++ != ')')
                    ierr = ARGER;
                    break;
                    }
                    setpiece = FALSE;
                }
            }

            break;

        
        case HA:            /* HALT or HANG */
        case ZHALT:     /* ZHALT */

            /* no arguments: HALT */
            if (*codptr == SP || *codptr == EOL || mcmnd == ZHALT) {

                if (mcmnd == ZHALT && *codptr != SP && *codptr != EOL) {
                    expr (STRING);
                    i = intexpr (argptr);
                
                    if (ierr > OK) break;
                }
                else {
                    halt:i = 0;
                }

                cleanup ();
                
                if (father) {           /* advertise death to parent *//* make sure father is waiting !!! */
                    if ((time (0L) - jobtime) < 120) sleep (2);
                    
                    kill (father, SIGUSR1);
                }

                exit (i);       /* terminate mumps */
            };
            /* with arguments: HANG */


        case HANG:          /* HANG */

            {
                unsigned long int waitsec;
                int millisec;
                
#ifdef USE_GETTIMEOFDAY
                struct timeval timebuffer;
#else
                struct timeb timebuffer;
#endif

                expr (STRING);
                numlit (argptr);

                if (ierr > OK) break;
                if (argptr[0] == '-') break;      /* negative values without effect */
                if (argptr[0] == '0') break;      /* zero without effect */
                
                waitsec = 0;
                millisec = 0;
                i = 0;
                
                for (;;) {       /* get integer and fractional part */ 
                    
                    if ((ch = argptr[i++]) == EOL) break;
                    
                    if (ch == '.') {
                        millisec = (argptr[i++] - '0') * 100;
                    
                        if ((ch = argptr[i++]) != EOL) {
                            millisec += (ch - '0') * 10;
        
                            if ((ch = argptr[i]) != EOL) {
                                millisec += (ch - '0');
                            }
                        }
                    
                        break;
                    }

                    waitsec = waitsec * 10 + ch - '0';
                }

                if ((i = waitsec) > 2) i -= 2;

#ifdef USE_GETTIMEOFDAY
                gettimeofday (&timebuffer, NULL);   /* get current time */

                waitsec += timebuffer.tv_sec;   /* calculate target time */
                millisec += timebuffer.tv_usec;
#else
                ftime (&timebuffer);    /* get current time */
                
                waitsec += timebuffer.time; /* calculate target time */
                millisec += timebuffer.millitm;
#endif

                if (millisec >= 1000) {
                    waitsec++;
                    millisec -= 1000;
                }

                /* do the bulk of the waiting with sleep() */
                while (i > 0) {
                    j = time (0L);
                    sleep ((unsigned) (i > 32767 ? 32767 : i)); /* sleep max. 2**15-1 sec */
                    i -= time (0L) - j; /* subtract actual sleeping time */

                    if (breakon && (ierr == INRPT)) goto err;

                    if (evt_async_enabled && (ierr == ASYNC)) goto err;
                }

                /* do the remainder of the waiting watching the clock */
                for (;;) {
                
#ifdef USE_GETTIMEOFDAY
                
                    gettimeofday (&timebuffer, NULL);
                
                    if (timebuffer.tv_sec > waitsec) break;
                    if (timebuffer.tv_sec == waitsec && timebuffer.tv_usec >= millisec) break;
#else
                    ftime (&timebuffer);

                    if (timebuffer.time > waitsec) break;
                    if (timebuffer.time == waitsec && timebuffer.millitm >= millisec) break;
#endif
                }
            }
            break;


        case HALT:          /* HALT */

            if (*codptr == SP || *codptr == EOL) goto halt;
            
            ierr = ARGLIST;
            break;


        case BREAK:

            if (*codptr == SP || *codptr == EOL) {
                
                if (breakon == FALSE) break;      /* ignore BREAK */
                
                if (usermode == 0) {
                    ierr = BKERR;
                    goto err;
                }
                
                zbflag = TRUE;
                ierr = OK - CTRLB;
                zb_entry:loadsw = TRUE;
                
#ifdef DEBUG_NEWSTACK
                printf ("CHECK 08 (Stack PUSH)\r\n");
#endif
                


                if (++nstx > NESTLEVLS) {
                    nstx--;
                    ierr = STKOV;
                
                    goto err;
                }
                else {
                    estack++;
                }

                nestc[nstx] = BREAK;

#ifdef DEBUG_NEWSTACK

                if (!cmdptr) printf ("CMDPTR is ZERO!\r\n");

#endif

                nestp[nstx] = cmdptr;   /* command stack address */
                nestn[nstx] = 0;    /*!!! save name */
                nestr[nstx] = roucur - rouptr;  /* save roucur */
                nestnew[nstx] = 0;
                ztrap[nstx][0] = EOL;
                nestlt[nstx] = level;
                level = 0;      /* save level */
                /* save BREAK information   */
                brkstk[nstx] = (((ECHOON ? 1 : 0) << 1) | test) << 3 | io;

                io = HOME;
                forsw = FALSE;
                cmdptr += stcpy (cmdptr, codptr) + 1;
                zerr = BKERR;
                goto restart;
            }

            expr (STRING);
            if (ierr > OK) break;

            switch (intexpr (argptr)) {
                
                case 2:
                    DSM2err = TRUE;
                    break;      /* enable DSM V 2 error processing */
                
                case -2:
                    DSM2err = FALSE;
                    break;      /* enable normal error processing  */
                
                case 0:
                    breakon = FALSE;
                    break;      /* disable CTRL/C */
                
                default:
                    breakon = TRUE;
                    break;      /* enable CTRL/C  */
            }
break;

        case VIEW:

            view_com ();

            if (repQUIT) {           /* VIEW 26: repeated QUIT action */
                
                while (repQUIT-- > 0) {

#ifdef DEBUG_NEWSTACK
                    printf ("CHECK 09 (Stack POP)\r\n");
#endif

                    if (nestc[nstx] == BREAK) {
                        if (repQUIT) continue;
                        ierr = OK - CTRLB;
                        
                        goto zgo;   /*cont. single step */
                    }

                    if (nestc[nstx] == FOR) {

                        stcpy (code, cmdptr = nestp[nstx--]);
                        estack--;
                        
                        codptr = code;
                        ftyp = fortyp[--forx];
                        fvar = forvar[forx];
                        finc = forinc[forx];
                        flim = forlim[forx];
                        fi = fori[forx];
                        
                        if (repQUIT) continue;
                        if ((forsw = (nestc[nstx] == FOR))) goto for_end;

                        goto next_line;
                    }

                    if (nestn[nstx]) {       /* reload routine */
                        namptr = nestn[nstx];
                        
                        if ((nestc[nstx] != XECUTE) || loadsw) {
                            stcpy (rou_name, namptr);
                            zload (rou_name);

                            ssvn_job_update ();
                            
                            dosave[0] = 0;
                        }

                        namptr--;
                    }

                    if (nestnew[nstx]) unnew ();   /* un-NEW variables */
                    
                    /* restore old pointers */
                    if ((mcmnd = nestc[nstx]) == BREAK) {
                        if (repQUIT) continue;
                        
                        goto restore;
                    }       /*cont. single step */
                    
                    if (mcmnd == DO_BLOCK) {
                        test = nestlt[nstx];
                        level--;
                    }                
                    else { /* pop $TEST */
                        level = nestlt[nstx];   /* pop level */
                    }

                    roucur = nestr[nstx] + rouptr;
                    stcpy (codptr = code, cmdptr = nestp[nstx--]);
                    estack--;
                    forsw = (nestc[nstx] == FOR);


                    loadsw = TRUE;

                    if (mcmnd == '$') {
                        if (repQUIT) return 0;
                        ierr = NOVAL;
                    }
                }
                repQUIT = 0;
            }
            break;

        /* Z-COMMANDS */
        case ZGO:

            /* ZGO with arguments: same as GOTO but with BREAK on */
            if (*codptr != EOL && *codptr != SP) {
                mcmnd = GOTO;
                zbflag = TRUE;
                ierr = OK - CTRLB;
                
                goto do_goto;
            }

            /* argumentless ZGO resume execution after BREAK */

            if (nestc[nstx] != BREAK) {
                ierr = LVLERR;
                break;
            }



            ierr = OK;      /* stop BREAKing */

zgo:

#ifdef DEBUG_NEWSTACK
            printf ("Zgoing: (Stack POP)\r\n");
#endif



            if (nestn[nstx]) {           /* reload routine */
                stcpy (rou_name, (namptr = nestn[nstx]));
                zload (rou_name);

                ssvn_job_update ();

                if (ierr > OK) break;
            }

            level = nestlt[nstx];
            roucur = nestr[nstx] + rouptr;
            io = brkstk[nstx];
            
            if (io & 020) {
                DSW &= ~BIT0;
            }
            else {
                DSW |= BIT0;        /* restore echo state */
            }

            test = (io & 010) >> 3; /* restore $TEST */

            /* restore $IO; default to HOME if channel not OPEN */
            if ((io &= 07) != HOME && devopen[io] == 0) io = HOME;
            
            stcpy (codptr = code, cmdptr = nestp[nstx--]);
            estack--;
            
            forsw = (nestc[nstx] == FOR);


            loadsw = TRUE;
            zbflag = FALSE;
            
            goto next0;


        case ZBREAK:

            if (*codptr == SP || *codptr == EOL) {
                ierr = ARGLIST;
                break;
            }

            expr (STRING);
            if (ierr > OK) break;

            zbreakon = tvexpr (argptr);
            if (hardcopy == DISABLE) set_zbreak (zbreakon ? STX : -1);   /* enable/disable CTRL/B */

            zbflag = FALSE;
            break;



    
        case ZLOAD:

            if (*codptr == EOL || *codptr == SP) {
                stcpy (varnam, rou_name);
            }
            else {
                expr (NAME);
            
                if (ierr > OK) break;
            
                codptr++;
            }

            dosave[0] = EOL;

            if (varnam[0] == EOL) {
                varerr[0] = EOL;
                ierr = NOPGM;
                break;
            }           /*error */

            loadsw = TRUE;

            /* a ZLOAD on the active routine always loads from disk */
            if (stcmp (varnam, rou_name) == 0) {                
                for (i = 0; i < NO_OF_RBUF; i++) {
            
                    if (rouptr == (buff + (i * PSIZE0))) {
                        pgms[i][0] = EOL;
            
                        break;
                    }
                }
            }

            zload (varnam);

            if (ierr > OK) break;          /* load file */

            stcpy (rou_name, varnam);
            ssvn_job_update ();

            break;

        case ZSAVE:

            if (*codptr == EOL || *codptr == SP) {

                if (rou_name[0] == EOL) {
                    varerr[0] = EOL;
                    ierr = NOPGM;

                    break;
                }           /*error */

                stcpy (varnam, rou_name);
            }
            else {
                expr (NAME);

                if (varnam[0] == '^') ierr = GLOBER;
                if (varnam[0] == '$') ierr = INVREF;
                if (ierr > OK) break;
                
                stcpy (rou_name, varnam);
                ssvn_job_update ();

                codptr++;
            }

            zsave (varnam);
            break;


        case ZREMOVE:

            {
                char *beg, *end;

                dosave[0] = EOL;

                if (*codptr == SP || *codptr == EOL) {           /* no args is ZREMOVE  all */
                    loadsw = TRUE;

                    for (i = 0; i < NO_OF_RBUF; i++) {

                        if (rouptr == buff + (i * PSIZE0)) {
                            pgms[i][0] = EOL;
                            break;
                        }

                    }

                    rouptr = buff + (i * PSIZE0);
                    rouend = rouins = rouptr;
                    roucur = buff + (NO_OF_RBUF * PSIZE0 + 1);
                    
                    *(rouptr) = EOL;
                    *(rouptr + 1) = EOL;
                    *(rouptr + 2) = EOL;
                    
                    argptr = partition;
                    rou_name[0] = EOL;

                    ssvn_job_update ();
                    
                    break;
                }
                if (*codptr == ':') {
                    beg = rouptr;
                }
                else if (*codptr == '*') {
                    beg = rouptr;
                
                    while ((end = (beg + UNSIGN (*beg) + 2)) < rouins) beg = end;
                
                    codptr++;
                }
                else {
                    lineref (&beg);
                    if (ierr > OK) break;
                }

                if ((end = beg) == 0) {
                    ierr = M13;
                    break;
                }

                if (*codptr == ':') {           /* same as above */
                    codptr++;
                
                    if (*codptr == '*') {
                        end = rouins;
                        codptr++;
                    }
                    else if (*codptr == ',' || *codptr == SP || *codptr == EOL) {
                        end = rouend;
                    }
                    else {
                        lineref (&end);
                        
                        if (end == 0) ierr = M13;
                        if (ierr > OK) break;
                
                        end = end + UNSIGN (*end) + 2;
                    }
                }
                else {
                    end = end + UNSIGN (*end) + 2;
                }

                if (beg < rouend) {           /* else there's nothing to zremove */

                    if (end >= rouend) {
                        end = rouend = beg;
                    }
                    else {
                        rouins = beg;

                        while (end <= rouend) *beg++ = (*end++);
                        
                        i = beg - end;
                        rouend += i;
                        
                        if (roucur > end) roucur += i;
                    }
                    
                    *end = EOL;
                    *(end + 1) = EOL;
                    
                    for (i = 0; i < NO_OF_RBUF; i++) {
                        if (rouptr == (buff + (i * PSIZE0))) {
                            ends[i] = rouend;
                            break;
                        }
                    }

                }
                break;
            }

        case ZINSERT:

            {
                char *beg;

                if (*codptr == EOL || *codptr == SP) {
                    ierr = ARGLIST;
                    break;
                }           /*error */
                
                dosave[0] = EOL;

                /* parse stringlit */
                expr (STRING);
                
                if (ierr > OK) break;
                
                if (*codptr != ':') {
                    zi (argptr, rouins);
                    break;
                }

                stcpy (tmp, argptr);
                codptr++;
                lineref (&beg);
                
                if (ierr > OK) break;      /* parse label */
                
                if (beg) {
                    beg = beg + UNSIGN (*beg) + 2;
                }
                else {
                    beg = rouptr;
                }
                
                if (beg > rouend + 1) {
                    ierr = M13;
                    break;
                }

                /* insert stuff */
                zi (tmp, beg);
                break;
            }


        /* PRINT is convenient -
        * but non-standard ZPRINT should be used instead */
        case 'p':

            if (standard) {
                ierr = NOSTAND;
                break;
            }


        case ZPRINT:
            
            {
                char *beg, *end;

                if (*codptr == SP || *codptr == EOL) {           /* no args is ZPRINT all */
                    beg = rouptr;
                    end = rouend;
                }
                else {
                    if (*codptr == ':') {
                        beg = rouptr;   /* from begin */
                    }
                    else if (*codptr == '*') {           /* from 'linepointer' */
                        beg = rouptr;
                        
                        while ((end = (beg + UNSIGN (*beg) + 2)) < rouins) beg = end;
                        codptr++;
                    }
                    else {
                        lineref (&beg);                        
                        if (ierr > OK) break;
                    }           /* line reference */
                    
                    if (beg == 0) {
                        beg = rouptr;
                        rouins = beg;
                        
                        if (*codptr != ':') break;
                    }

                    if (*codptr == ':') {
                        codptr++;   /* to end */
                    
                        if (*codptr == SP || *codptr == ',' || *codptr == EOL)
                            end = rouend;
                        else {
                            if (*codptr == '*') {
                                end = rouins;
                                codptr++;
                            }               
                            else {  /* to 'linepointer' */
                                lineref (&end);
                    
                                if (ierr > OK) break;  /* line reference */
                                end = end + UNSIGN (*end) + 2;
                            }
                        }
                    }
                    else {
                        end = beg + 1;
                    }
                }
                if (rouend < end) end = rouend - 1;

                for (; beg < end; beg += UNSIGN (*beg) + 2) {
                    
                    if (crlf[io]) {
                        write_m ("\012\201");
                    }
                    else {
                        write_m ("\012\015\201");
                    }
                    
                    if ((*(beg + 1)) == EOL) break;
                    
                    write_m (beg + 1);
                    if (ierr > OK) break;
                }
                
                rouins = beg;
            }

            if (crlf[io]) {
                write_m ("\012\201");
            }
            else {
                write_m ("\012\015\201");
            }

            break;

        case ZWATCH:
            {
                char op;


                if (((ch = *codptr) == SP) || ch == EOL) {
                    
                    set_io(UNIX);                        
                    
                    if (dbg_enable_watch) {
                        printf ("Watchpoints disabled.\n"); 
                        dbg_enable_watch = 0;
                    }
                    else {
                        printf ("Watchpoints enabled.\n");
                        dbg_enable_watch = 1;
                    }
                    
                    break;        

                }

                if ((ch = *codptr) == '(') {
                    ierr = ARGLIST;
                    goto err;
                }

                for (;;) {

                    switch (ch) {
                        
                        case '?':
                        case '+':
                        case '-':
                            op = ch;
                            codptr++;
                            break;

                        default: 
                            ierr = ARGLIST;
                            goto err;
                    }

                    expr (NAME);                            /* try to interpret an mname */
                    
                    if (ierr > OK) goto err;

                    stcpy (vn, varnam);

                    switch (op) {
                        
                        case '+':
                            dbg_add_watch (vn);
                            break;

                        case '-':
                            dbg_remove_watch (vn);
                            break;

                        case '?':
                            dbg_dump_watch (vn);
                            break;

                    }                

                    if (ierr > OK) goto err;

                    if ((ch = *(codptr + 1)) == EOL) {
                        codptr++;
                        break;
                    }
                    else if ((ch = *(codptr + 1)) == ',') {
                        codptr += 2;
                        ch = *codptr;
                    }
                    else {
                        ierr = ARGLIST;
                        goto err;
                    }
                } 
                

                break;
            }


        case ZASSERT:
            {
                expr (STRING);

                if (ierr > OK) goto err;

                if (tvexpr (argptr) == 0) {
                    ierr = ASSERT;
                    goto err;
                }

                break;
            }

        case ZWRITE:


zwrite:
            {
                short k;
                char w_tmp[512];
                char zwmode;


                if (io != HOME && devopen[io] == 'r') {
                    ierr = NOWRITE;
                    goto err;
                }

                tmp3[0] = SP;
                tmp3[1] = EOL;
                
                if ((ch = (*codptr)) == '(') {    /* exclusive zwrite */

                    for (;;) {
                        
                        codptr++;
                        expr (NAME);
                        
                        if (ierr > OK) goto err;
                        if (varnam[0] == '^') {
                            ierr = GLOBER;
                            goto err;
                        }

                        i = 0;

                        while (varnam[i] != EOL) {
                            
                            if (varnam[i] == DELIM) {
                                ierr = SBSCR;
                                goto err;
                            }

                            i++;
                        }

                        if (stcat (tmp3, varnam) == 0) {
                            ierr = MXSTR;
                            goto err;
                        }

                        if (stcat (tmp3, " \201") == 0) {
                            ierr = MXSTR;
                            goto err;
                        }

                        if ((ch = *++codptr) == ')') {
                            codptr++;
                            break;
                        }

                        if (ch != ',') {
                            ierr = COMMAER;
                            goto err;
                        }
                    }
                }
                else {
                    if (ch != SP && ch != EOL) goto zwritep;
                }

                /* no arguments: write local symbol table. */
                stcpy (tmp, " $\201");

                for (;;) {
                    ordercnt = 1L;
                    
                    symtab (bigquery, &tmp[1], tmp2);
                    
                    if (*tmp2 == EOL || ierr == INRPT) break;
                    w_tmp[0] = '=';

                    /* subscripts: internal format different from external one */
                    k = 0;
                    i = 1;
                    j = 0;
                    
                    while ((ch = tmp2[k++]) != EOL) {
                        
                        if (ch == '"') {
                            
                            if (j && tmp2[k] == ch) {
                                k++;
                            }
                            else {
                                toggle (j);
                                continue;
                            }

                        }
                        
                        if (j == 0) {
                    
                            if (ch == '(' || ch == ',') {
                                tmp[i++] = DELIM;
                                
                                continue;
                            }
                    
                            if (ch == ')') break;
                        }
                    
                        tmp[i++] = ch;
                    }
                    
                    tmp[i] = EOL;
                    if (kill_ok (tmp3, tmp) == 0) continue;

                    write_m (tmp2);
                    symtab (get_sym, &tmp[1], &w_tmp[1]);
                    write_m (w_tmp);
                    write_m ("\012\015\201");
                }
                
                break;

zwritep:

                expr (NAME);

                //if (varnam[0] == '^') ierr = GLOBER;
                if (ierr > OK) goto err;
                
                codptr++;

                if (varnam[0] == '$') {
                
                    if ((varnam[1] | 0140) == 'z' && (varnam[2] | 0140) == 'f') {
                        w_tmp[0] = '$';
                        w_tmp[1] = 'Z';
                        w_tmp[2] = 'F';
                        w_tmp[3] = '(';
                
                        for (i = 0; i < 44; i++) {

                            if (zfunkey[i][0] != EOL) {
                                intstr (&w_tmp[4], i + 1);
                                stcat (w_tmp, ")=\201");
                                write_m (w_tmp);
                                write_m (zfunkey[i]);
                                write_m ("\012\015\201");
                            }

                        }

                        break;
                    }
                    else {
                        break;      /* do not zwrite special variables etc. other than $ZF */
                    }
                }
                
                if (varnam[0] != '^') {
                    symtab (dat, varnam, tmp2);
                    zwmode = 'L';
                }
                else {
                    if (varnam[1] == '$') {
                        ssvn (dat, varnam, tmp2);
                        zwmode = '$';
                    }
                    else {
                        global (dat, varnam, tmp2);
                        zwmode = '^';
                    }
                }

                if (tmp2[0] == '0') break;      /* variable not defined */
                
                /* if $D(@varnam)=10 get next entry */
                if (tmp2[1] == '0') {
                    ordercnt = 1L;

                    if (varnam[0] != '^') {
                        symtab (query, varnam, tmp2);
                        zwmode = 'L';
                    }
                    else {
                        if (varnam[1] == '$') {
                            ssvn (query, varnam, tmp2);
                            zwmode = '$';
                        }
                        else {
                            global (query, varnam, tmp2);
                            zwmode = '^';
                        }
                    }
                }
                else {
                    k = 0;
                    i = 0;
                    j = 0;
                
                    while ((ch = varnam[k++]) != EOL) {
                        
                        if (ch == DELIM) {
                            
                            if (j) {
                                tmp2[i++] = '"';
                                tmp2[i++] = ',';
                                tmp2[i++] = '"';

                                continue;
                            }

                            j++;
                
                            tmp2[i++] = '(';
                            tmp2[i++] = '"';
                
                            continue;
                        }
                
                        if ((tmp2[i++] = ch) == '"')
                        tmp2[i++] = ch;
                    }
                
                    if (j) {
                        tmp[i++] = '"';
                        tmp2[i++] = ')';
                    }
                
                    tmp2[i] = EOL;
                }
                
                for (;;) {           /* subscripts: internal format different from external one */
                    k = 0;
                    i = 0;
                    j = 0;

                    while ((ch = tmp2[k++]) != EOL) {

                        if (ch == '"') {
                            if (j && tmp2[k] == ch)
                                k++;
                            else {
                                toggle (j);
                                continue;
                            }
                        }
                        
                        if (j == 0) {
                            
                            if (ch == '(' || ch == ',') {
                                tmp[i++] = DELIM;
                        
                                continue;
                            }
                        
                            if (ch == ')') break;
                        }

                        tmp[i++] = ch;
                    }

                    tmp[i] = EOL;
                    i = 0;
                    
                    while (tmp[i] == varnam[i]) {

                        if (varnam[i] == EOL) break;
                    
                        i++;
                    }

                    if (varnam[i] != EOL) break;
                    if (tmp[i] != EOL && tmp[i] != DELIM) break;
                    
                    tmp3[0] = EOL;

                    switch (zwmode) {

                        case 'L':
                            symtab (dat, tmp, tmp3);
                            symtab (get_sym, tmp, &w_tmp[1]);

                            break;


                        case '$':
                            ssvn (dat, tmp, tmp3);
                            ssvn (get_sym, tmp, &w_tmp[1]);

                            break;


                        case '^':
                            global (dat, tmp, tmp3);
                            global (get_sym, tmp, &w_tmp[1]);

                            break;
                    }

                    if (tmp3[0] != '0' && tmp3[1] != '0') {                        
                        
                        write_m (tmp2);
                        
                        w_tmp[0] = '=';
                        
                        write_m (w_tmp);
                        write_m ("\012\015\201");

                    }
                    
                    ordercnt = 1L;
                    
                    switch (zwmode) {

                        case 'L':
                            symtab (query, tmp, tmp2);

                            break;


                        case '$':
                            ssvn (query, tmp, tmp2);

                            break;


                        case '^':
                            global (query, tmp, tmp2);

                            break;

                    }
                    
                    if (ierr == INRPT) break;
                }

                break;
            }


        case ZTRAP:

            if (*codptr == SP || *codptr == EOL) {
                ierr = ZTERR;
                varnam[0] = EOL;
                
                break;
            }

            expr (NAME);
            stcpy (varerr, varnam);

            if (ierr) break;

            if (*++codptr == ':') {           /* parse postcond */
                codptr++;

                expr (STRING);

                if (ierr > OK) goto err;

                if (tvexpr (argptr) == FALSE) break;
            }

            ierr = ZTERR;
            break;


        case ZALLOCATE:

            /* argumentless is not permitted */
            if (*codptr == SP || *codptr == EOL) {
                ierr = ARGLIST;
                break;
            }

            expr (NAME);
            
            if (ierr > OK) goto err;
            
            tmp[0] = SP;
            stcpy (&tmp[1], varnam);
            stcat (tmp, "\001\201");

            timeout = (-1L);        /* no timeout */
            
            if (*++codptr == ':') {
                codptr++;
            
                expr (STRING);
            
                timeout = intexpr (argptr);
            
                if (ierr > OK) goto err;
                if (timeout < 0L) timeout = 0L;
            }

            lock (tmp, timeout, ZALLOCATE);
            break;
            

        case ZDEALLOCATE:

            tmp[0] = SP;
            
            if (*codptr == SP || *codptr == EOL) {
                tmp[1] = EOL;
            }
            else {
                expr (NAME);
            
                if (ierr > OK) goto err;
            
                stcpy (&tmp[1], varnam);
            
                codptr++;
            }

            lock (tmp, -1L, ZDEALLOCATE);   /* -1: no timeout */
            break;

            /* user defined Z-COMMAND */


        case PRIVATE:

private:            /* for in-MUMPS defined commands */
            i = 0;
            j = 0;
            ch = 0;

            while ((tmp2[i] = *codptr) != EOL) {
                
                if (tmp2[i] == SP && !j) {
                    tmp2[i] = EOL;
                    break;
                }

                if (tmp2[i] == '"') j = (!j);
                
                if (!j) {
                    
                    if (tmp2[i] == '(') ch++;
                    if (tmp2[i] == ')') ch--;
                
                    if (!ch && tmp2[i] == ',') {       /* next argument: */
                
                        tmp2[i] = EOL;  /* call afterwards again */
                        i = 0;
                
                        while (tmp3[i] != EOL) i++;

                        j = i;
                        ch = 1;
                
                        while (ch < i) tmp3[j++] = tmp3[ch++];
                
                        tmp3[j - 1] = SP;
                        tmp3[j] = EOL;
                
                        codptr++;
                
                        j = 0;
                        ch = 0;
                
                        break;
                    }
                }
                
                i++;
                codptr++;            
            }

            if (j || ch) {
                ierr = INVREF;
                goto err;
            }

            stcat (tmp3, codptr);
            stcpy (code, "d ^%\201");
            stcat (code, &tmp3[1]);

            codptr = code;
            privflag = TRUE;

            goto next_cmnd;

evthandler:            /* for event handlers */
            i = 0;
            j = 0;
            ch = 0;

            while ((tmp2[i] = *codptr) != EOL) {
                
                if (tmp2[i] == SP && !j) {
                    tmp2[i] = EOL;
                    break;
                }

                if (tmp2[i] == '"') j = (!j);
                
                if (!j) {
                    
                    if (tmp2[i] == '(') ch++;
                    if (tmp2[i] == ')') ch--;
                
                    if (!ch && tmp2[i] == ',') {       /* next argument: */
                
                        tmp2[i] = EOL;  /* call afterwards again */
                        i = 0;
                
                        while (tmp3[i] != EOL) i++;

                        j = i;
                        ch = 1;
                
                        while (ch < i) tmp3[j++] = tmp3[ch++];
                
                        tmp3[j - 1] = SP;
                        tmp3[j] = EOL;
                
                        codptr++;
                
                        j = 0;
                        ch = 0;
                
                        break;
                    }
                }
                
                i++;
                codptr++;            
            }

            if (j || ch) {
                ierr = INVREF;
                goto err;
            }

            stcpy (code, "d \201");
            stcat (code, tmp3);

            codptr = code;
            privflag = TRUE;

            goto next_cmnd;            

        case ABLOCK:
        case AUNBLOCK:
            {

                /* declare and initialize table of events to be blocked/unblocked with this command */
                short evt_mask[EVT_MAX];
                
                for (i = 0; i < EVT_MAX; i++) evt_mask[i] = 0;


                /* argumentless ABLOCK/AUNBLOCK: block/unblock everything */
                if (((ch = *codptr) == SP) || ch == EOL) {
                    
                    for (i = 0; i < EVT_MAX; i++) evt_mask[i] = 1;

                }
                else if (*codptr == '(') {
                    /* exclusive ABLOCK/AUNBLOCK */

                    short evt_exclusions[EVT_MAX];
                    
                    codptr++;


                    for (i = 0; i < EVT_MAX; i++) evt_exclusions[i] = FALSE;

                    for (;;) {

                        expr (STRING);

                        if (ierr == BRAER) ierr = OK;
                        if (ierr > OK) goto err;

                        codptr++;

                        stcpy (vn, argptr);

                        if (stcmp (vn, "COMM\201") == 0) {
                            evt_exclusions[EVT_CLS_COMM] = TRUE;
                        }
                        else if (stcmp (vn, "HALT\201") == 0) {
                            evt_exclusions[EVT_CLS_HALT] = TRUE;
                        }
                        else if (stcmp (vn, "IPC\201") == 0) {
                            evt_exclusions[EVT_CLS_IPC] = TRUE;
                        }
                        else if (stcmp (vn, "INTERRUPT\201") == 0) {
                            evt_exclusions[EVT_CLS_INTERRUPT] = TRUE;
                        }
                        else if (stcmp (vn, "POWER\201") == 0) {
                            evt_exclusions[EVT_CLS_POWER] = TRUE;
                        }
                        else if (stcmp (vn, "TIMER\201") == 0) {
                            evt_exclusions[EVT_CLS_TIMER] = TRUE;
                        }
                        else if (stcmp (vn, "USER\201") == 0) {
                            evt_exclusions[EVT_CLS_USER] = TRUE;
                        }
                        else if (stcmp (vn, "WAPI\201") == 0) {
                            evt_exclusions[EVT_CLS_WAPI] = TRUE;
                        }
                        else {
                            ierr = CMMND;
                            goto err;
                        }

                        if ((ch = *(codptr + 1)) == EOL || ch == SP) {
                            codptr++;
                            break;
                        }                       
                        if ((ch = *(codptr + 1)) == ')') {
                            codptr++;
                            break;
                        }

                    }
                
                    for (i = 0; i < EVT_MAX; i++) {

                        if (evt_exclusions[i] == FALSE) evt_mask[i] = 1;

                    }

                }
                else {
                    /* inclusive ABLOCK/AUNBLOCK */

                    for (;;) {

                        expr (STRING);                            /* try to interpret a string */                        
                        if (ierr > OK) goto err;

                        codptr++;

                        stcpy (vn, argptr);

                        if (stcmp (vn, "COMM\201") == 0) {
                            evt_mask[EVT_CLS_COMM] = 1;
                        }
                        else if (stcmp (vn, "HALT\201") == 0) {
                            evt_mask[EVT_CLS_HALT] = 1;
                        }
                        else if (stcmp (vn, "IPC\201") == 0) {
                            evt_mask[EVT_CLS_IPC] = 1;
                        }
                        else if (stcmp (vn, "INTERRUPT\201") == 0) {
                            evt_mask[EVT_CLS_INTERRUPT] = 1;
                        }
                        else if (stcmp (vn, "POWER\201") == 0) {
                            evt_mask[EVT_CLS_POWER] = 1;
                        }
                        else if (stcmp (vn, "TIMER\201") == 0) {
                            evt_mask[EVT_CLS_TIMER] = 1;
                        }
                        else if (stcmp (vn, "USER\201") == 0) {
                            evt_mask[EVT_CLS_USER] = 1;
                        }
                        else if (stcmp (vn, "WAPI\201") == 0) {
                            evt_mask[EVT_CLS_WAPI] = 1;
                        }
                        else {
                            ierr = CMMND;
                            goto err;
                        }
                    
                        if (ierr > OK) goto err;


                        if ((ch = *(codptr)) == EOL || ch == SP) {                            
                            break;
                        }

                    } 

                }

                for (i = 0; i < EVT_MAX; i++) {                    

                    if (evt_mask[i] > 0) {

                        if (mcmnd == ABLOCK) {
                            evt_ablock (i);
                        }
                        else {
                            evt_aunblock (i);
                        }
                    }

                }
            
                
                break;
            }


        case ASSIGN:
            ierr = CMMND;
            break;


        case ASTOP:
        case ASTART:
            {

                /* declare and initialize table of events to be enabled with this command */
                short evt_mask[EVT_MAX];
                short new_status;

                if (mcmnd == ASTART) {
                    new_status = EVT_S_ASYNC;
                }
                else {
                    new_status = EVT_S_DISABLED;
                }


                for (i = 0; i < EVT_MAX; i++) evt_mask[i] = EVT_S_NOMODIFY;


                /* argumentless ASTART/ASTOP: enable/disable everything */
                if (((ch = *codptr) == SP) || ch == EOL) {
                    
                    for (i = 0; i < EVT_MAX; i++) evt_mask[i] = new_status;

                }
                else if (*codptr == '(') {
                    /* exclusive ASTART */
                    
                    short evt_exclusions[EVT_MAX];

                    codptr++;

                    for (i = 0; i < EVT_MAX; i++) evt_exclusions[i] = FALSE;

                    for (;;) {

                        expr (STRING);

                        if (ierr == BRAER) ierr = OK;
                        if (ierr > OK) goto err;

                        codptr++;

                        stcpy (vn, argptr);

                        if (stcmp (vn, "COMM\201") == 0) {
                            evt_exclusions[EVT_CLS_COMM] = TRUE;
                        }
                        else if (stcmp (vn, "HALT\201") == 0) {
                            evt_exclusions[EVT_CLS_HALT] = TRUE;
                        }
                        else if (stcmp (vn, "IPC\201") == 0) {
                            evt_exclusions[EVT_CLS_IPC] = TRUE;
                        }
                        else if (stcmp (vn, "INTERRUPT\201") == 0) {
                            evt_exclusions[EVT_CLS_INTERRUPT] = TRUE;
                        }
                        else if (stcmp (vn, "POWER\201") == 0) {
                            evt_exclusions[EVT_CLS_POWER] = TRUE;
                        }
                        else if (stcmp (vn, "TIMER\201") == 0) {
                            evt_exclusions[EVT_CLS_TIMER] = TRUE;
                        }
                        else if (stcmp (vn, "USER\201") == 0) {
                            evt_exclusions[EVT_CLS_USER] = TRUE;
                        }
                        else if (stcmp (vn, "WAPI\201") == 0) {
                            evt_exclusions[EVT_CLS_WAPI] = TRUE;
                        }
                        else {
                            ierr = CMMND;
                            goto err;
                        }

                        if ((ch = *(codptr + 1)) == EOL || ch == SP) {
                            codptr++;
                            break;
                        }                       
                        if ((ch = *(codptr + 1)) == ')') {
                            codptr++;
                            break;
                        }

                    }
                
                    for (i = 0; i < EVT_MAX; i++) {

                        if (evt_exclusions[i] == FALSE) evt_mask[i] = new_status;

                    }

                }
                else {
                    /* inclusive ASTART */

                    for (;;) {

                        expr (STRING);                            /* try to interpret a string */                        
                        if (ierr > OK) goto err;

                        codptr++;

                        stcpy (vn, argptr);

                        if (stcmp (vn, "COMM\201") == 0) {
                            evt_mask[EVT_CLS_COMM] = new_status;
                        }
                        else if (stcmp (vn, "HALT\201") == 0) {
                            evt_mask[EVT_CLS_HALT] = new_status;
                        }
                        else if (stcmp (vn, "IPC\201") == 0) {
                            evt_mask[EVT_CLS_IPC] = new_status;
                        }
                        else if (stcmp (vn, "INTERRUPT\201") == 0) {
                            evt_mask[EVT_CLS_INTERRUPT] = new_status;
                        }
                        else if (stcmp (vn, "POWER\201") == 0) {
                            evt_mask[EVT_CLS_POWER] = new_status;
                        }
                        else if (stcmp (vn, "TIMER\201") == 0) {
                            evt_mask[EVT_CLS_TIMER] = new_status;
                        }
                        else if (stcmp (vn, "USER\201") == 0) {
                            evt_mask[EVT_CLS_USER] = new_status;
                        }
                        else if (stcmp (vn, "WAPI\201") == 0) {
                            evt_mask[EVT_CLS_WAPI] = new_status;
                        }
                        else {
                            ierr = CMMND;
                            goto err;
                        }
                    
                        if (ierr > OK) goto err;


                        if ((ch = *(codptr)) == EOL || ch == SP) {                            
                            break;
                        }

                    } 

                }

                for (i = 0; i < EVT_MAX; i++) {                    

                    if (evt_status[i] == EVT_S_SYNC && evt_mask[i] == EVT_S_ASYNC) {
                        
                        /* cannot enable both synchronous and asynchronous 
                           event processing on the same event class at the
                           same time */

                        ierr = M102;
                        goto err;

                    }
                    else {

                        if (evt_mask[i] > EVT_S_NOMODIFY) {
                            evt_status[i] = evt_mask[i];
                        }
                        
                    }

                }
            
                if (mcmnd == ASTART)  {
                    evt_async_enabled = TRUE;
                }
                else {                    
                    short disabled_evt_count = 0;

                    for (i = 0; i < EVT_MAX; i++) {
                        if (evt_status[i] == EVT_S_DISABLED) {
                            disabled_evt_count++;
                        }
                    }

                    if (disabled_evt_count == (EVT_MAX - 1)) evt_async_enabled = FALSE;

                }

                break;
            }




        case ETRIGGER:

            ierr = CMMND;
            break;
        
        
        case ESTART:
            {
                /* stack ^$EVENT */
                symtab (new_sym, "^$EVENT\201", "\201");

                break;
            }
        
        
        case ESTOP:
        default:
            ierr = CMMND;
    
    }               /* command switch */

    if ((ch = *codptr) == EOL) {
        if (ierr) goto err;
        if (forsw) goto for_end;
        
        mcmnd = 0;
        
        goto next_line;
    }

    if (ch == SP) {
        if (ierr == OK) goto next0;

        goto err;
    }

    if (ch != ',' && ierr == OK) { 
        ierr = SPACER;
    }
    else if (ierr <= OK) {
        if (*++codptr != SP && *codptr != EOL) goto again;

        ierr = ARGLIST;
    }
    
    /* else goto err; */

/* error */
err:

    /*
     * ierr == ASYNC means that the previous command was interrupted by
     * an async event. It is not a real error, so just go on to the next
     * command after resetting ierr = OK. 
     */
    if (ierr == ASYNC) {
        ierr = OK;
        goto next_cmnd;
    }

    if (ierr < 0) {
        
        ierr += CTRLB;

        if (ierr == OK) {
            zbflag = TRUE;
            
            goto zb_entry;
        }
    }

    if (ierr > OK ) {

        char er_buf[ERRLEN];
        
        merr_set_ecode_ierr ();

        stcpy (er_buf, errmes[ierr]);
        stcnv_m2c (er_buf);

        m_log (LOG_ERR, er_buf);
    }

    if (ierr != INRPT && promflag == FALSE) {
        prompt[0] = '"';

        stcpy (&prompt[1], defprompt);
        stcat (prompt, "\"\201");
    }

    zerr = ierr;
    ierr = OK;
    /*     goto restart;    */


restart:
    
    if (param) goto restore;

    dosave[0] = EOL;
    setpiece = FALSE;
    setop = 0;
    privflag = FALSE;

    if (ierr == INRPT) goto err;
    if (zerr == STORE) symtab (kill_all, "", "");

    if (errfunlvl > 0) {
        errfunlvl--;
    }
    else {

        if (zerr == OK) {
            zerror[0] = EOL;    /* reset error */
        }
        else {

#ifdef DEBUG_STACK
            printf ("Storing NESTERR\r\n");
#endif

            nesterr = nstx; /* save stack information at error */

            for (i = 1; i <= nstx; i++) getraddress (callerr[i], i);

            zerror[0] = '<';

            if (etxtflag) {
                stcpy (&zerror[1], errmes[zerr]);
            }
            else {
                intstr (&zerror[1], zerr);
            }

            stcat (zerror, ">\201");

            if (rou_name[0] != EOL) {
                char *j0;
                char *j1;
                char tmp1[256];


            
                if (nestc[nstx] == XECUTE) {
                    
                    if (nestn[nstx]) {       /* reload routine */
                        zload (nestn[nstx]);
                        ierr = OK;
                    }
                
                    roucur = nestr[nstx] + rouptr;  /* restore roucur */
                }


            
                j0 = (rouptr - 1);
                j = 0;
                tmp1[0] = EOL;
                j0++;
                
                if (roucur < rouend) {
                    
                    while (j0 < (roucur - 1)) {

                        j1 = j0++;
                        j++;

                        if ((*j0 != TAB) && (*j0 != SP)) {
                        
                            j = 0;
                        
                            while ((tmp1[j] = (*(j0++))) > SP) {
                        
                                if (tmp1[j] == '(') tmp1[j] = EOL;
                        
                                j++;
                            }
                        
                            tmp1[j] = EOL;
                            j = 0;
                        }
                        
                        j0 = j1;
                        j0 += (UNSIGN (*j1)) + 2;
                    }
                }

                stcat (zerror, tmp1);

                if (j > 0) {
                    i = stlen (zerror);
                    zerror[i++] = '+';
                    
                    intstr (&zerror[i], j);
                }

                stcat (zerror, "^\201");
            

            
                if (nestc[nstx] == XECUTE) {
                    
                    if (nestn[nstx]) {       /* reload routine */
                        zload (rou_name);

                        ssvn_job_update ();

                        ierr = OK;
                    }
                
                    stcat (zerror, nestn[nstx]);
                }
                else
                    stcat (zerror, rou_name);
            }

            if (zerr == UNDEF) zerr = M6;

            /* undefined: report variable name */
            if (zerr == UNDEF || zerr == SBSCR || zerr == NAKED || zerr == ZTERR || zerr == DBDGD || zerr == LBLUNDEF || zerr == NOPGM || zerr == M6 || zerr == M7 || zerr == M13) { 

                int f;      /* include erroneous reference */

                f = stlen (zerror);
                zerror[f++] = SP;
                zname (&zerror[f], varerr);
            }           /* end varnam section */
        }
    }

    roucur = buff + (NO_OF_RBUF * PSIZE0 + 1);
    tmp4[0] = EOL;

    while (ierr != (OK - CTRLB)) {


        /* standard error handling */
        if (etrap[0] != EOL && stcmp (ecode, "") != 0)  {

            on_frame_entry ();
            
            /* disable $ZTRAP error handling */
            ztrap[nstx][0] = EOL;

            stcpy (tmp4, etrap);
            stcat (tmp4, " quit:$quit \"\" quit\201");

            if (etrap_lvl > 1) {
                /* we've encountered an error within an error handler. 
                   save off the error code at merr_stack[nstx + 1].ECODE */

                stcpy (merr_stack[nstx + 1].ECODE, ecode);
                merr_topstk = nstx + 1;
                etrap_lvl++;

            }
            else {
                merr_topstk = nstx;
                etrap_lvl++;
            }

            break;

        }



        if (ztrap[nstx][0] != EOL && !DSM2err) {           

#ifdef DEBUG_NEWSTACK

            printf ("Dropped into Ztrap [");
            
            for (loop = 0; loop < 20 && ztrap[nstx][loop] != EOL; loop++) {
                printf ("%c", ztrap[nstx][loop]);
            }

            printf ("]\r\n");

#endif

            tmp4[0] = GOTO;
            tmp4[1] = SP;
            stcpy (&tmp4[2], ztrap[nstx]);
            ztrap[nstx][0] = EOL;

#ifdef DEBUG_NEWSTACK

            printf ("Set tmp4 to [");
            for (loop = 0; tmp4[loop] != EOL; loop++) printf ("%c", tmp4[loop]);            
            printf ("]\r\n");

#endif

            break;
        }

        if (nstx == 0) {

#ifdef DEBUG_NEWSTACK
            printf ("Nestx was Zero\r\n");
#endif
        
            forx = 0;
            cmdptr = cmdstack;
            namptr = namstck;
            level = 0;
            errfunlvl = 0;
            io = HOME;      /* trap to direct mode: USE 0 */
            
            if (zerr == INRPT && filter) {
                tmp4[0] = 'h';
                tmp4[1] = EOL;
            }

            if (DSM2err && (ztrap[NESTLEVLS + 1][0] != EOL)) {           /* DSM V.2 error trapping */

#ifdef DEBUG_NEWSTACK
                printf ("Ztrap 2\r\n");
#endif

                tmp4[0] = GOTO;
                tmp4[1] = SP;   /* GOTO errorhandling */
                
                stcpy (&tmp4[2], ztrap[NESTLEVLS + 1]);
                ztrap[NESTLEVLS + 1][0] = EOL;
                
            }

            break;
        }

#ifdef DEBUG_NEWSTACK
        printf ("Nestc[nstx] is [%d]\r\n", nestc[nstx]);
#endif

        if (nestc[nstx] == BREAK) break;

        if (nestc[nstx] == FOR) {
            ftyp = fortyp[--forx];
            fvar = forvar[forx];
            finc = forinc[forx];
            flim = forlim[forx];
            fi = fori[forx];
        }
        else {

            if (nestc[nstx] == DO_BLOCK) {
                test = nestlt[nstx];
                level--;
            }
            else { /* pop $TEST */
                level = nestlt[nstx];   /* pop level */
            }

#ifdef DEBUG_NEWSTACK
            printf ("Nestn[nstx] is [%d]\r\n", nestn[nstx]);
#endif

            if (nestn[nstx]) {           /* 'reload' routine */
                namptr = nestn[nstx];
                stcpy (rou_name, namptr);
                zload (rou_name);

                ssvn_job_update ();
                
                dosave[0] = 0;
                
                namptr--;
            }

#ifdef DEBUG_NEWSTACK
            printf ("Execcing the rest...\r\n");
#endif

            roucur = nestr[nstx] + rouptr;

            if (nestnew[nstx]) unnew ();       /* un-NEW variables */
            
            cmdptr = nestp[nstx];

            if (nestc[nstx--] == '$') {           /* extrinsic function/variable */
                *argptr = EOL;
                ierr = zerr;
                errfunlvl++;
                
                return 0;
            }
            estack--;
        }
    }

    forsw = FALSE;

    /* PRINTING ERROR MESSAGES */
    if (tmp4[0] == EOL) {
        
        if (zerr == BKERR && brkaction[0] != EOL) {
            stcpy (code, brkaction);
            codptr = code;

            if (libcall == TRUE) {
                return zerr;
            } 
            else {
                goto next_cmnd;
            }
        }

        if (libcall == TRUE) return zerr;

        DSW &= ~BIT0;       /* enable ECHO */

        // print here
        {
            char *t_rtn;
            char *t_nsn = (char *) malloc (STRLEN * sizeof (char));
            char *t_cod;
            int t_pos;

            t_rtn = strtok (zerror, ">");
            t_rtn = strtok (NULL, ">");

            if (t_rtn[1] == '%') {
                strcpy (t_nsn, "SYSTEM");
            }
            else {
                strcpy (t_nsn, nsname);
            }

            if (deferred_ierr > OK) {
                t_cod = deferrable_code;
                t_pos = deferrable_codptr - code + 3;
            }
            else {
                t_cod = code;
                t_pos = codptr - code + 3;
            }
      
            merr_dump (zerr, t_rtn, t_nsn, t_cod, t_pos);

            free (t_nsn);
        }


    }
    else {            
        stcpy (code, tmp4);
        
        codptr = code;
        tmp4[0] = EOL;

        goto next_cmnd;
    }

restore:

    io = HOME;
    codptr = code;

    if (param > 0) {

        j = 0;
        ch = 0;
        paramx++;
        param--;
        
        for (;;) {
            if (m_argv[++j][0] == '-') {
                i = 0;
            
                while ((m_argv[j][++i] != 0) && (m_argv[j][i] != 'x'));
                
                if (m_argv[j][i] != 'x') continue;
            
                j++;
            
                if (++ch < paramx) continue;
            
                strcpy (code, m_argv[j]);
                break;
            }
            else {
                if (++ch < paramx) continue;
            
                strcpy (code, "d ");
                strcpy (&code[2], m_argv[j]);
                break;
            }
        }
        code[strlen (code)] = EOL;
        codptr = code;
        goto next_cmnd;

    }

    if (usermode == 0) {               /* application mode: direct mode implies HALT */
        code[0] = 'H';
        code[1] = EOL;
        codptr = code;

        goto next_cmnd;
    }

    if (libcall == TRUE) {             /* library mode: don't go to direct mode, just return */
        return ierr;
    }


    do {

        if (filter == FALSE && promflag) {
#ifndef HAVE_LIBREADLINE                
            stcpy (code, "w \201");
            stcpy (&code[2], prompt);
#else
            stcpy (code, " \201");
            stcpy (&code[2], " \201");                
#endif                
            promflag = FALSE;
        }
        else {


            if (dbg_enable_watch && dbg_pending_watches) dbg_dump_watchlist ();

            /* DIRECT-MODE PROMPT HERE */
#if defined(HAVE_LIBREADLINE) && !defined(_AIX)
            {
                char *fmrl_buf;
                char fmrl_prompt[256];
                HIST_ENTRY **hist_list;                
                int hist_idx;
                HIST_ENTRY *hist_ent;

                if (quiet_mode == FALSE) {
                    if (tp_level == 0) {
                        snprintf (fmrl_prompt, 256, "\r\n%s> ", nsname);
                    }
                    else {
                        snprintf (fmrl_prompt, 256, "\r\nTL%d:%s> ", tp_level, nsname);
                    }
                }
                set_io (UNIX);

                /* readline() does its own malloc() */
                fmrl_buf = readline (fmrl_prompt);

                if (!fmrl_buf) {
                    set_io (UNIX);
                    printf ("\n");
                    set_io (MUMPS);
                    
                    goto halt;
                }
                
                if (strlen (fmrl_buf) > 0) {
                    add_history (fmrl_buf);
                }                    

                if (strcmp (fmrl_buf, "?") == 0) {

                    /* Invoke Online Help */

                    set_io (MUMPS);
                    stcpy (code, "DO ^%ZHELP\201");

                }
                else if (strcmp (fmrl_buf, "events") == 0) {

                    char stat_desc[30];
                    char *evclass_name;

                    printf ("\n%-20s %-15s %s\n", "Event Class", "Processing Mode", "ABLOCK Count");
                    printf ("%-20s %-15s %s\n", "-----------", "---------------", "------------");

                    for (i = 0; i < EVT_MAX; i++) {

                        evclass_name = evt_class_name_c (i);

                        switch (evt_status[i]) {
                            case EVT_S_DISABLED:
                                strcpy (stat_desc, "Disabled");
                                break;
                            case EVT_S_ASYNC:
                                strcpy (stat_desc, "Asynchronous");
                                break;
                            case EVT_S_SYNC:
                                strcpy (stat_desc, "Synchronous");
                        }

                        printf ("%-20s %-15s %d\n", evclass_name, stat_desc, evt_blocks[i]);

                    }
                    
                    
                }
                else if (strcmp (fmrl_buf, "wh") == 0) {
                    write_history (history_file);
                }
                else if (strcmp (fmrl_buf, "tdump") == 0) {
                    tp_tdump();
                }
                else if (isdigit(fmrl_buf[0]) || (fmrl_buf[0] == '(') || (fmrl_buf[0] == '-') || (fmrl_buf[0] == '\'') || (fmrl_buf[0] == '+')) {

                    snprintf (code, 255, "W %s", fmrl_buf);
                    stcnv_c2m (code);

                    set_io (MUMPS);

                }
#if !defined(__APPLE__)
                else if (strcmp (fmrl_buf, "history") == 0) {

                    /* History List */                                                

                    hist_list = history_list ();
                    if (hist_list) {

                        for (i = 0; hist_list[i]; i++) {
                            printf("%d: %s\n", i + history_base, hist_list[i]->line);
                        }
                        
                    }

                    stcpy (code, " \201");

                    set_io (MUMPS);

                }
#endif                    
                else if (strncmp (fmrl_buf, "rcl", 3) == 0) {

                    /* Recall History Item */
                    
                                            
                    
                    if (!isdigit (fmrl_buf[4])) {
                        fprintf (stderr, "invalid history index '%s'\n", &fmrl_buf[4]);

                        set_io (MUMPS);
                        stcpy (code, " \201");

                        break;
                    }

                    hist_idx = atoi (&fmrl_buf[4]);

                    if ((hist_idx > history_length) || (hist_idx < 1)) {
                        fprintf (stderr, "history entry %d out of range (valid entries are 1-%d)\n", hist_idx, history_length);

                        set_io (MUMPS);
                        stcpy (code, " \201");

                        break;
                    }                        

                    hist_ent = history_get (hist_idx);

                    printf ("%s\n", hist_ent->line);

                    strncpy (code, hist_ent->line, 255);
                    stcnv_c2m (code);

                    set_io (MUMPS);

                }
                else {

                    /* Pass to M Interpreter */

                    set_io (MUMPS);

                    strncpy (code, fmrl_buf, 255);
                    stcnv_c2m (code);

                }

                /* free the buffer malloc()'d by readline() */
                free (fmrl_buf);                    
            }
#else

#ifdef __EMSCRIPTEN__

            fprintf (stderr, "xecline():  at direct-mode prompt\n");
            fgets (code, 255, stdin);
            stcnv_c2m (code);

#else
            {
                char fmrl_prompt[256];

                if (tp_level == 0) {
                    snprintf (fmrl_prompt, 256, "\r\n%s> \201", nsname);
                }
                else {
                    snprintf (fmrl_prompt, 256, "\r\nTL%d:%s> \201", tp_level, nsname);
                }
                
                write_m (fmrl_prompt);

                read_m (code, -1L, 0, 255); /* Not necessarily STRLEN? */
            }

#endif

#endif
            promflag = TRUE;

            if (ierr > OK) goto err;


            if (code[0] == EOL && zbflag && nestc[nstx] == BREAK) {

                ierr = OK - CTRLB;
                goto zgo;
            }           /* single step */
        }
    } 
    while (code[0] == EOL);

    if (promflag) write_m ("\r\n\201");

    /* automatic ZI in direct mode: insert an entry with TAB */
    i = (-1);
    j = 0;
    ierr = OK;

    while (code[++i] != EOL) {
        if (code[i] == '"') toggle (j);

        if (code[i] == TAB && j == 0) {
            dosave[0] = EOL;

            zi (code, rouins);
            if (ierr) goto err;
            goto restore;
        }
    }

    code[++i] = EOL;
    code[++i] = EOL;

    roucur = buff + (NO_OF_RBUF * PSIZE0 + 1);

    goto next_cmnd;

skip_line:

    if (forsw) goto for_end;
    goto next_line;

}               /*end of xecline() */

void on_frame_entry(void)
{
    
    char *placebuf;
    char pb[256];

    //write_m("in on_frame_entry()\201");

    stcpy (merr_stack[nstx].ECODE, ecode);
    stcpy (merr_stack[nstx].MCODE, code);

    stcpy (pb, zerror);
    stcnv_m2c (pb);

    placebuf = strtok (pb, ">");
    placebuf = strtok (NULL, ">");
                
    if (placebuf != NULL) {
        stcnv_c2m (placebuf);
        
        stcpy (merr_stack[nstx].PLACE, placebuf);
    }
    else {
        stcpy (merr_stack[nstx].PLACE, "\201");
    }

}
